from relationalai.semantics.metamodel.compiler import Pass
from relationalai.semantics.metamodel import ir, builtins as rel_builtins, factory as f, visitor
from relationalai.semantics.metamodel import helpers
from relationalai.semantics.metamodel.util import FrozenOrderedSet

from relationalai.semantics.lqp.utils import output_names

from typing import List, Sequence, Tuple, Union

# Deduplicate Vars in Updates and Outputs.
class DeduplicateVars(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        r = self.VarDeduplicator()
        return r.walk(model)

    # Return 1) a new list of Values with no duplicates (at the object level) and
    # 2) equalities between any original Value and a deduplicated Value.
    @staticmethod
    def dedup_values(vals: Sequence[ir.Value]) -> Tuple[List[ir.Value], List[ir.Lookup]]:
        # If a var is seen more than once, it is a duplicate and we will create
        # a new Var and equate it with the seen one.
        seen_vars = set()

        new_vals = []
        eqs = []

        for i, val in enumerate(vals):
            # Duplicates can only occur within Vars.
            # TODO: we don't know for sure if these are the only relevant cases.
            if isinstance(val, ir.Default) or isinstance(val, ir.Var):
                var = val if isinstance(val, ir.Var) else val.var
                if var in seen_vars:
                    new_var = ir.Var(var.type, var.name + "_dup_" + str(i))
                    new_val = new_var if isinstance(val, ir.Var) else ir.Default(new_var, val.value)
                    new_vals.append(new_val)
                    eqs.append(f.lookup(rel_builtins.eq, [new_var, var]))
                else:
                    seen_vars.add(var)
                    new_vals.append(val)
            else:
                # No possibility of problematic duplication.
                new_vals.append(val)

        return new_vals, eqs

    # Returns a reconstructed output with no duplicate variable objects
    # (dedup_values) and now necessary equalities between any two previously
    # duplicate variables.
    @staticmethod
    def dedup_output(output: ir.Output) -> List[Union[ir.Output, ir.Lookup]]:
        vals = helpers.output_values(output.aliases)
        deduped_vals, req_lookups = DeduplicateVars.dedup_values(vals)
        # Need the names so we can recombine.
        alias_names = output_names(output.aliases)
        new_output = ir.Output(
            output.engine,
            FrozenOrderedSet(list(zip(alias_names, deduped_vals))),
            output.keys,
            output.annotations,
        )
        return req_lookups + [new_output]

    # Returns a replacement update with no duplicate variable objects
    # (dedup_values) and now necessary equalities between any two previously
    # duplicate variables.
    @staticmethod
    def dedup_update(update: ir.Update) -> List[Union[ir.Update, ir.Lookup]]:
        deduped_vals, req_lookups = DeduplicateVars.dedup_values(update.args)
        new_update = ir.Update(
            update.engine,
            update.relation,
            tuple(deduped_vals),
            update.effect,
            update.annotations,
        )
        return req_lookups + [new_update]

    # Does the actual work.
    class VarDeduplicator(visitor.Rewriter):
        def __init__(self):
            super().__init__()

        # We implement handle_logical instead of handle_update/handle_output
        # because in addition to modifying said update/output we require new
        # lookups (equality between original and deduplicated variables).
        def handle_logical(self, node: ir.Logical, parent: ir.Node):
            # In order to recurse over subtasks.
            node = super().handle_logical(node, parent)

            new_body = []
            for subtask in node.body:
                if isinstance(subtask, ir.Output):
                    new_body.extend(DeduplicateVars.dedup_output(subtask))
                elif isinstance(subtask, ir.Update):
                    new_body.extend(DeduplicateVars.dedup_update(subtask))
                else:
                    new_body.append(subtask)

            return ir.Logical(
                node.engine,
                node.hoisted,
                tuple(new_body),
                node.annotations
            )



from __future__ import annotations

from relationalai.semantics.metamodel import ir, factory as f, helpers
from relationalai.semantics.metamodel.util import OrderedSet, ordered_set
from relationalai.semantics.lqp.rewrite.flatten import Flatten, negate, extend_body
from relationalai.semantics.lqp.algorithms import is_script, mk_assign

class FlattenScript(Flatten):
    """
    Flattens Match nodes inside @script Sequence blocks. This pass extends Flatten
    of standard Logicals, to reuse a number of utilities, but DOES NOT flatten Match
    nodes outside of scripts (which are handled by the Flatten pass).

    Unlike the regular Flatten pass which extracts to top-level, FlattenScript
    maintains order by inserting intermediate relations right before they're used
    within the Sequence. This is necessary because order matters in a script Sequence.

    Additionally, if the original Logical is a Loopy instruction, the flattened Logicals are
    made Loopy too (`@assign`).

    Example:

    === BEFORE ===
    Logical
        Sequence @script @algorithm
            Logical
                dom(n)
                Match ⇑[k]
                    Logical ⇑[k]
                        value(n, k)
                    Logical ⇑[k]
                        k = 0
                → derive result(n, k) @assign @global
                filter(n)

    === AFTER ===
    Logical
        Sequence @script @algorithm
            Logical
                dom(n)
                filter(n)
                Logical ⇑[v]
                    value(n, v)
                → derive _match_1(n, v) @assign
            Logical
                dom(n)
                filter(n)
                Logical ⇑[v]
                    v = 0
                Not
                    _match_1(n, _)
                → derive _match_2(n, v) @assign
            Logical
                dom(n)
                filter(n)
                Union ⇑[v]
                    _match_1(n, v)
                    _match_2(n, v)
                → derive result(n, v) @assign @global
    """

    class Context(Flatten.Context):
        """Extended context with script tracking."""
        def __init__(self, model: ir.Model, options: dict):
            super().__init__(model, options)
            self.in_script: bool = False

    def rewrite(self, model: ir.Model, options: dict = {}) -> ir.Model:
        """Traverse the model and flatten Match nodes inside script Sequences."""
        ctx = FlattenScript.Context(model, options)
        result = self.handle(model.root, ctx)

        if result.replacement is None:
            return model

        # Convert relations list to FrozenOrderedSet (adding any new intermediate relations)
        new_relations = OrderedSet.from_iterable(model.relations).update(ctx.rewrite_ctx.relations).frozen()

        return ir.Model(
            model.engines,
            new_relations,
            model.types,
            result.replacement
        )

    def handle(self, task: ir.Task, ctx: Flatten.Context) -> Flatten.HandleResult:
        """Override handle to add Loop support."""
        if isinstance(task, ir.Loop):
            return self.handle_loop(task, ctx)
        return super().handle(task, ctx)

    def handle_loop(self, task: ir.Loop, ctx: Flatten.Context) -> Flatten.HandleResult:
        """Recursively handle the body of the loop."""
        result = self.handle(task.body, ctx)

        assert(result.replacement)

        # If body unchanged, return original loop
        if result.replacement is task.body:
            return Flatten.HandleResult(task)

        # Return new loop with handled body
        return Flatten.HandleResult(ir.Loop(
            task.engine,
            task.hoisted,
            task.iter,
            result.replacement,
            task.concurrency,
            task.annotations
        ))

    def handle_logical(self, task: ir.Logical, ctx: Context):  # type: ignore[override]
        """
        Handle Logical nodes.

        Outside scripts: simple traversal to find nested scripts.
        Inside scripts: prevent extraction to top-level and keep everything in sequence.

        Note: Type checker complains about parameter type narrowing, but this is safe
        because we only create FlattenScript.Context in our own rewrite() method.
        """
        # Recursively process children
        body: OrderedSet[ir.Task] = ordered_set()
        for child in task.body:
            result = self.handle(child, ctx)
            if result.replacement is not None:
                if ctx.in_script and isinstance(result.replacement, ir.Logical) and not result.replacement.hoisted:
                    # Inside script: inline simple logicals without hoisting
                    body.update(result.replacement.body)
                else:
                    body.add(result.replacement)

        if not body:
            return Flatten.HandleResult(None)

        return Flatten.HandleResult(
            ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations)
        )

    def flatten_match_in_logical(self, logical: ir.Logical, match: ir.Match, match_idx: int, ctx: Context) -> list[ir.Logical]:
        """
        Flatten a Match inside a Logical within a script Sequence.
        Returns a list of Logicals to be inserted in sequence.
        """
        if not match.tasks:
            return [logical]

        # Split the logical into: tasks_before, match, tasks_after
        tasks_before = list(logical.body[:match_idx])
        tasks_after = list(logical.body[match_idx + 1:])

        # Separate tasks_after into filters (non-Update tasks) and updates (Update tasks)
        # Filters are constraints that should be included in all branches
        filters = [task for task in tasks_after if not isinstance(task, ir.Update)]
        updates = [task for task in tasks_after if isinstance(task, ir.Update)]

        # Compute exposed variables
        exposed_vars = self.compute_exposed_vars(match, match.tasks, ctx)

        # Use dependency analysis for branch bodies (like flatten.py does)
        branch_dependencies = ctx.info.task_dependencies(match)

        # Collect all dependencies for the final Logical (tasks before + filters after)
        final_dependencies = tasks_before + filters

        # Negation length for wildcards
        outputs = ctx.info.task_outputs(match)
        negation_len = len(outputs) if outputs else 0

        # Result: list of Logicals to insert
        result_logicals = []
        references = []
        negated_reference = None

        # Process each branch
        for branch in match.tasks:
            # Create connection relation for this branch
            name = helpers.create_task_name(self.name_cache, branch, "_match")
            relation = helpers.create_connection_relation(branch, exposed_vars, ctx.rewrite_ctx, name)

            # Handle the branch (recursively process nested structures)
            result = self.handle(branch, ctx)
            branch_content = result.replacement if result.replacement else branch

            # Update dependency tracking if branch was transformed
            if result.replacement:
                ctx.info.replaced(branch, result.replacement)

            # Build logical for this branch
            branch_body: OrderedSet[ir.Task] = ordered_set()

            # Add dependencies (using dependency analysis, not all tasks)
            branch_body.update(branch_dependencies)

            # Add branch content using extend_body helper
            extend_body(branch_body, branch_content)

            # Add negation of previous branches (after branch content)
            if negated_reference:
                branch_body.add(negated_reference)

            # Add derive to connection relation
            branch_update = f.derive(relation, exposed_vars)
            branch_body.add(branch_update)

            # Create the Logical for this branch
            branch_logical = mk_assign(ir.Logical(match.engine, tuple(), tuple(branch_body)))

            result_logicals.append(branch_logical)

            # Update references for final union
            reference = f.lookup(relation, exposed_vars)
            negated_reference = negate(reference, negation_len)
            references.append(reference)

        # Create final Logical with Union and remaining tasks
        final_body: OrderedSet[ir.Task] = ordered_set()
        final_body.update(final_dependencies)

        # Add union of all branches
        union = f.union(references, match.hoisted)
        final_body.add(union)

        # Add updates that came after the match (filters are already in dependencies)
        final_body.update(updates)

        # Create final logical preserving the original annotations
        final_logical = ir.Logical(logical.engine, logical.hoisted, tuple(final_body), logical.annotations)

        result_logicals.append(final_logical)

        return result_logicals

    def handle_sequence(self, task: ir.Sequence, ctx: Context):  # type: ignore[override]
        """
        Handle a Sequence.

        If it's a script: set context flag and flatten Match nodes within.
        If not a script: simple traversal (Flatten pass already processed it).

        Note: Type checker complains about parameter type narrowing, but this is safe
        because we only create FlattenScript.Context in our own rewrite() method.
        """
        if not is_script(task):
            # Not a script sequence - already processed by Flatten, just return as-is
            return Flatten.HandleResult(task)

        # This is a script - mark context and process with flattening
        old_in_script = ctx.in_script
        ctx.in_script = True

        # Process the sequence tasks
        new_tasks: list[ir.Task] = []
        for child in task.tasks:
            # Check if this child is a Logical with Match that needs flattening
            if isinstance(child, ir.Logical):
                new_tasks.extend(self.try_flatten_logical(child, ctx))
                continue

            # No flattening needed, process normally
            result = self.handle(child, ctx)
            if result.replacement is not None:
                new_tasks.append(result.replacement)

        # Restore context
        ctx.in_script = old_in_script

        return Flatten.HandleResult(
            ir.Sequence(task.engine, task.hoisted, tuple(new_tasks), task.annotations)
        )

    def try_flatten_logical(self, logical: ir.Logical, ctx: Context) -> list[ir.Logical]:
        """
        Flatten all Matches in a Logical.
        Iteratively flattens until no more Matches remain in any of the resulting Logicals.
        """
        worklist = [logical]
        result = []

        while worklist:
            current = worklist.pop()

            # Find first Match in current logical
            match = None
            match_idx = -1
            for i, child in enumerate(current.body):
                if isinstance(child, ir.Match):
                    match = child
                    match_idx = i
                    break

            if match is None:
                # No Match found - this logical is done
                result.append(current)
            else:
                # Flatten and add results back to worklist for further processing.
                # Reverse so that pop() returns them in the original order.
                flattened = self.flatten_match_in_logical(current, match, match_idx, ctx)
                worklist.extend(reversed(flattened))

        return result



from relationalai.semantics.metamodel.compiler import Pass
from relationalai.semantics.metamodel import ir, builtins as rel_builtins, factory as f, visitor
from relationalai.semantics.metamodel.typer import typer

from typing import List, Sequence, Tuple, Union

# Rewrite constants to vars in Updates. This results in a more normalized format where
# updates contain only variables. This allows for easier rewrites in later passes.
class ConstantsToVars(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        r = self.ConstantToVarRewriter()
        return r.walk(model)

    # Return 1) a new list of Values with no duplicates (at the object level) and
    # 2) equalities between any original Value and a deduplicated Value.
    @staticmethod
    def replace_constants_with_vars(vals: Sequence[ir.Value]) -> Tuple[List[ir.Value], List[ir.Lookup]]:
        new_vals = []
        eqs = []

        for i, val in enumerate(vals):
            if isinstance(val, ir.PyValue) or isinstance(val, ir.Literal):
                # Replace constant with a new Var.
                typ = typer.to_type(val)
                assert isinstance(typ, ir.ScalarType), "can only replace scalar constants with vars"
                new_var = ir.Var(typ, f"{typ.name.lower()}_{i}")
                new_vals.append(new_var)
                eqs.append(f.lookup(rel_builtins.eq, [new_var, val]))
            else:
                new_vals.append(val)

        return new_vals, eqs

    @staticmethod
    def dedup_update(update: ir.Update) -> List[Union[ir.Update, ir.Lookup]]:
        deduped_vals, req_lookups = ConstantsToVars.replace_constants_with_vars(update.args)
        new_update = ir.Update(
            update.engine,
            update.relation,
            tuple(deduped_vals),
            update.effect,
            update.annotations,
        )
        return req_lookups + [new_update]

    # Does the actual work.
    class ConstantToVarRewriter(visitor.Rewriter):
        def __init__(self):
            super().__init__()

        # We implement handle_logical instead of handle_update because in
        # addition to modifying said update we require new lookups (equality
        # between original and deduplicated variables).
        def handle_logical(self, node: ir.Logical, parent: ir.Node):
            # In order to recurse over subtasks.
            node = super().handle_logical(node, parent)

            new_body = []
            for subtask in node.body:
                if isinstance(subtask, ir.Update):
                    new_body.extend(ConstantsToVars.dedup_update(subtask))
                else:
                    new_body.append(subtask)

            return ir.Logical(
                node.engine,
                node.hoisted,
                tuple(new_body),
                node.annotations
            )



import warnings

from relationalai.semantics.lqp import lqp_bundle, model2lqp, result_helpers, utils

ir = lqp_bundle.ir

__all__ = ['ir', 'model2lqp', 'result_helpers', 'utils']

warnings.warn(
    "relationalai.early_access.lqp.* is deprecated. "
    "Please migrate to relationalai.semantics.lqp.*",
    DeprecationWarning,
    stacklevel=2,
)



#pyright: reportPrivateImportUsage=false
from __future__ import annotations
import io
import json
import os
import re
import sys
import requests
import rich
import click
import functools
import pytz
from typing import NoReturn

from relationalai.util.constants import TOP_LEVEL_PROFILE_NAME
from relationalai.errors import RAIException
from rich.table import Table
from typing import Callable, Dict, Any, List, cast
from ..clients import config
from click.core import Context
from rich.console import Console
from rich import box as rich_box
from collections import defaultdict
from packaging.version import Version
from ..clients.config import ConfigFile
from datetime import datetime, timedelta
from click.formatting import HelpFormatter
from ..clients.client import ResourcesBase
from relationalai.tools.constants import GlobalProfile, SHOW_FULL_TRACES
from relationalai.tools.cli_controls import divider
from relationalai.util.format import humanized_bytes, humanized_duration
from InquirerPy.base.control import Choice

#--------------------------------------------------
# Helpers
#--------------------------------------------------

@functools.cache
def get_config(profile:str|Dict[str, Any]|None = None):
    return config.Config(profile or GlobalProfile.get())

@functools.cache
def get_resource_provider(platform:str|None=None, _cfg:config.Config|None = None) -> ResourcesBase:
    cfg = _cfg or get_config()
    platform = platform or cfg.get("platform", "snowflake")
    if platform == "snowflake":
        from relationalai.clients.resources.snowflake.cli_resources import CLIResources
        provider = CLIResources(config=cfg)
    elif platform == "azure":
        from relationalai.clients.resources.azure.azure import Resources
        provider = Resources(config=cfg)
    else:
        from .. import Resources
        provider = Resources(config=cfg)
    return provider

def unexpand_user_path(path):
    """Inverse of os.path.expanduser"""
    home_dir = os.path.expanduser('~')
    if path.startswith(home_dir):
        return '~' + path[len(home_dir):]
    return path

def account_from_url(account_or_url:str):
    regex = r"https://app.snowflake.com/([^/]+)/([^/]+)/?.*"
    match = re.match(regex, account_or_url)
    if match:
        org = match.group(1)
        account = match.group(2)
        return f"{org}-{account}"
    elif "app.snowflake.com" in account_or_url or "https://" in account_or_url:
        raise ValueError("URL not of the form https://app.snowflake.com/[org]/[account]")
    else:
        return account_or_url

def supports_platform(*platform_names: str):
    def decorator(cmd: click.Command):
        setattr(cmd, "__available_platforms", platform_names)
        cb = cmd.callback
        assert cb
        def checked(*args, **kwargs):
            assert cmd.name
            assert_command_available(cmd.name, command_available(cmd))
            return cb(*args, **kwargs)

        cmd.callback = checked
        return cmd
    return decorator

def command_available(cmd: click.Command) -> bool:
    available_platforms = getattr(cmd, "__available_platforms", ())
    platform = get_config().get("platform", "")
    return not available_platforms or not platform or platform in available_platforms

def assert_command_available(name: str, available: bool, plural=False):
    if not available:
        platform = get_config().get("platform", "")
        rich.print(f"[yellow]{name} {'are' if plural else 'is'} not available for {platform}")
        divider()
        sys.exit(1)

def coming_soon():
    rich.print("[yellow]This isn't quite ready yet, but it's coming soon!")
    divider()
    sys.exit(1)

def issue_top_level_profile_warning():
    rich.print("[yellow]Warning: Using a top-level profile is not recommended")
    rich.print("[yellow]Consider naming the profile by adding \\[profile.<name>] to your raiconfig.toml file\n")
    rich.print("[yellow]Example:")
    rich.print("[yellow]\\[profile.default]")
    rich.print("[yellow]platform = \"snowflake\"")
    rich.print("[yellow]account = ...")
    divider()

def ensure_config(profile:str|None=None) -> config.Config:
    cfg = get_config(profile)
    if not cfg.file_path:
        rich.print("[yellow bold]No configuration file found.")
        rich.print("To create one, run: [green bold]rai init[/green bold]")
        divider()
        sys.exit(1)
    return cfg

def latest_version(package_name):
    """Get the current version of a package on PyPI."""
    url = f"https://pypi.org/pypi/{package_name}/json"
    response = requests.get(url)
    if response.status_code == 200:
        data = response.json()
        version = data['info']['version']
        return version
    else:
        return None

def is_latest_cli_version():
    from .. import __version__
    latest_ver_str = latest_version("relationalai")
    latest_ver = Version(latest_ver_str) if latest_ver_str else Version("0.0.0")
    version = Version(__version__)
    return version >= latest_ver, version, latest_ver

#--------------------------------------------------
# Validation
#--------------------------------------------------

EMPTY_STRING_REGEX = re.compile(r'^\S+$')
ENGINE_NAME_REGEX = re.compile(r'^[a-zA-Z]\w{2,}$')
COMPUTE_POOL_REGEX = re.compile(r'^[a-zA-Z][a-zA-Z0-9_]*$')
PASSCODE_REGEX = re.compile(r'^(\d{6})?$')
ENGINE_NAME_ERROR = "Min 3 chars, start with letter, only letters, numbers, underscores allowed."
UUID = re.compile('[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}')

def validate_engine_name(name:str) -> tuple[bool, str|None]:
    if not ENGINE_NAME_REGEX.match(name):
        return False, ENGINE_NAME_ERROR
    return True, None

#--------------------------------------------------
# Engine types & selection helpers (Snowflake)
#--------------------------------------------------

def format_state_with_color(state: str) -> str:
    """Format engine state with colors for display."""
    if not state:
        return ""
    state_upper = state.upper()
    if state_upper == "READY":
        return f"[green]{state_upper}[/green]"
    if state_upper == "SUSPENDED":
        return f"[yellow]{state_upper}[/yellow]"
    if state_upper in ("PENDING", "SYNCING", "PROCESSING"):
        return f"[bold yellow]{state_upper}[/bold yellow]"
    if state_upper in ("ABORTED", "QUARANTINED", "GONE"):
        return f"[red]{state_upper}[/red]"
    return state_upper

def _get_engine_type_api():
    # Local import to avoid importing snowflake modules for non-snowflake usage
    from relationalai.clients.resources.snowflake import EngineType
    return EngineType

def _get_internal_engine_sizes() -> list[str]:
    # Local import to avoid snowflake imports for non-snowflake usage
    from relationalai.clients.resources.snowflake import INTERNAL_ENGINE_SIZES
    return list(INTERNAL_ENGINE_SIZES)

def get_engine_type_choices(cfg: config.Config, exclude_types: List[str] | None = None) -> List[Choice]:
    """Get sorted list of engine type choices for interactive selection."""
    EngineType = _get_engine_type_api()
    if exclude_types is None:
        exclude_types = []
    exclude_types_upper = [et.upper() for et in exclude_types]
    engine_types_list = [EngineType.LOGIC, EngineType.ML, EngineType.SOLVER]
    engine_types_list = [et for et in engine_types_list if et.upper() not in exclude_types_upper]
    engine_types_list.sort(key=lambda et: EngineType.get_label(et))
    return [
        Choice(value=et, name=f"{EngineType.get_label(et)}: {EngineType.get_description(et)}")
        for et in engine_types_list
    ]

def select_engine_type_interactive(cfg: config.Config) -> str:
    """Show interactive engine type selection and return the selected type."""
    from . import cli_controls as controls
    rich.print("")
    engine_type_choices = get_engine_type_choices(cfg)
    return controls.select("Engine type:", cast("list[str | Choice]", engine_type_choices), None, newline=True)

def select_engine_interactive(
    provider: ResourcesBase,
    prompt: str = "Select an engine:",
    engine_name: str | None = None,
    engines: List[Dict[str, Any]] | None = None,
) -> tuple[str, str | None] | None:
    """Interactive engine picker returning (name, type)."""
    from . import cli_controls as controls

    engine_map: Dict[str, tuple[str, str | None]] = {}

    def get_engines():
        engine_list = engines if engines is not None else provider.list_engines()
        engine_map.clear()
        items: List[str] = []
        EngineType = _get_engine_type_api()
        for engine in engine_list:
            eng_name = engine.get("name", "")
            if engine_name and eng_name.upper() != engine_name.upper():
                continue
            eng_type = engine.get("type", "")
            eng_size = engine.get("size", "")
            if eng_type:
                label = f"{EngineType.get_label(eng_type)} ({eng_type})" if EngineType.is_valid(eng_type) else f"{eng_type} ({eng_type})"
                display = f"{eng_name}, {label}, {eng_size}"
            else:
                display = f"{eng_name}, {eng_size}" if eng_size else eng_name
            engine_map[display] = (eng_name, eng_type or None)
            items.append(display)
        return items

    # Auto-select when engine_name uniquely identifies an engine
    if engine_name:
        engine_list = engines if engines is not None else provider.list_engines()
        matches = [e for e in engine_list if e.get("name", "").upper() == engine_name.upper()]
        if len(matches) == 1:
            e = matches[0]
            return (e.get("name", ""), e.get("type"))
        if len(matches) == 0:
            return None

    selected = controls.fuzzy_with_refetch(prompt, "engines", get_engines)
    if not selected or isinstance(selected, Exception):
        return None
    return engine_map.get(selected)

def select_engine_with_state_filter(
    provider: ResourcesBase,
    engine_name: str | None,
    engine_type: str | None,
    state_filter: str,
    prompt_no_name: str,
    prompt_with_name: str,
    error_no_engines: str,
    error_no_matching: str,
) -> tuple[str, str | None] | None:
    """Select an engine with optional state filtering + optional name filtering."""
    EngineType = _get_engine_type_api()

    if not engine_name:
        filtered = provider.list_engines(state_filter)
        if not filtered:
            exit_with_error(error_no_engines)
        return select_engine_interactive(provider, prompt_no_name, engines=filtered)

    # If type provided and valid, return directly
    if engine_type and EngineType.is_valid(engine_type):
        return (engine_name, engine_type)

    # Filter by name + state; selection handles (name,type)
    filtered = provider.list_engines(state_filter, name=engine_name)
    if not filtered:
        exit_with_error(error_no_matching)
    return select_engine_interactive(provider, prompt_with_name, engine_name=engine_name, engines=filtered)

def ensure_engine_type_for_snowflake(
    provider: ResourcesBase,
    engine_name: str,
    engine_type: str | None,
    error_message: str,
) -> str:
    """Ensure engine_type is provided and valid; default to LOGIC if omitted."""
    EngineType = _get_engine_type_api()
    # If --type was omitted, default to LOGIC for backwards compatibility
    if engine_type is None:
        return EngineType.LOGIC
    assert isinstance(engine_type, str)
    if engine_type == "" or not EngineType.is_valid(engine_type):
        cfg = get_config()
        if engine_type == "":
            rich.print(f"[yellow]Empty engine type provided for engine '{engine_name}'.")
        else:
            rich.print(f"[yellow]Invalid engine type '{engine_type}' for engine '{engine_name}'.")
        return select_engine_type_interactive(cfg)
    return engine_type

def build_engine_operation_messages(
    provider: ResourcesBase,
    engine_name: str,
    engine_type: str | None,
    action: str,
    action_past: str,
) -> tuple[str, str]:
    EngineType = _get_engine_type_api()
    if engine_type:
        label = EngineType.get_label(engine_type) if EngineType.is_valid(engine_type) else engine_type
        return (f"{action} {label} engine '{engine_name}'", f"{label} Engine '{engine_name}' {action_past.lower()}")
    return (f"{action} '{engine_name}' engine", f"Engine '{engine_name}' {action_past.lower()}")

def prompt_and_validate_engine_name(name: str | None) -> str:
    """Prompt for engine name if missing; validate using ENGINE_NAME_REGEX."""
    from . import cli_controls as controls
    if not name:
        name = controls.prompt(
            "Engine name:",
            name,
            validator=ENGINE_NAME_REGEX.match,
            invalid_message=ENGINE_NAME_ERROR,
            newline=True,
        )
    assert isinstance(name, str)
    return name

def validate_auto_suspend_mins(auto_suspend_mins: int | str | None) -> int | None:
    if auto_suspend_mins is None:
        return None
    if isinstance(auto_suspend_mins, int):
        return auto_suspend_mins
    error_msg = f"[yellow]Error: auto_suspend_mins must be an integer instead of {type(auto_suspend_mins)}"
    try:
        return int(auto_suspend_mins)
    except ValueError:
        exit_with_error(error_msg)
    return None

def get_engine_type_for_creation(provider: ResourcesBase, cfg: config.Config, engine_type: str | None) -> str | None:
    """Get engine type for engine creation; defaults to LOGIC when omitted."""
    EngineType = _get_engine_type_api()
    if engine_type is None:
        return EngineType.LOGIC
    if engine_type == "" or not EngineType.is_valid(engine_type):
        if engine_type == "":
            rich.print("[yellow]Empty engine type provided.")
        else:
            valid_types_display = ", ".join(EngineType.get_all_types())
            rich.print(f"[yellow]Invalid engine type '{engine_type}'. Valid types: {valid_types_display}")
        return select_engine_type_interactive(cfg)
    return engine_type

def get_and_validate_engine_size(
    provider: ResourcesBase,
    cfg: config.Config,
    size: str | None,
    engine_type: str | None = None,
) -> str:
    from . import cli_controls as controls
    EngineType = _get_engine_type_api()
    internal_sizes = set(_get_internal_engine_sizes())

    cloud_provider = provider.get_cloud_provider()
    valid_sizes = provider.get_engine_sizes(cloud_provider)

    # Engine-type-aware filtering:
    # Internal sizes (XS/S/M/L) are only valid for LOGIC engines (and only on some accounts).
    # For ML/SOLVER, hide them to avoid presenting invalid options.
    if engine_type and EngineType.is_valid(engine_type) and engine_type != EngineType.LOGIC:
        valid_sizes = [s for s in valid_sizes if s not in internal_sizes]

    # Ask if missing and not in config
    if not size and not cfg.get("engine_size", None):
        rich.print("")
        # This refers to the cloud backing your Snowflake account (AWS/Azure), not the engine "platform".
        size = controls.fuzzy(f"Engine size (Snowflake cloud: {cloud_provider.upper()}):", choices=valid_sizes)
    elif size is None and cfg.get("engine_size", None):
        size = cfg.get("engine_size", None)
    if not isinstance(size, str) or size not in valid_sizes:
        exit_with_error(f"\nInvalid engine size [yellow]{size}[/yellow] provided. Please check your config.\n\nValid sizes: [green]{valid_sizes}[/green]")
    return size

def create_engine_with_spinner(
    provider: ResourcesBase,
    engine_name: str,
    engine_size: str,
    engine_type: str | None,
    auto_suspend_mins: int | None,
) -> None:
    """Create an engine with appropriate spinner messages and error handling."""
    from .cli_controls import Spinner

    EngineType = _get_engine_type_api()
    # Build creation message with engine type when available
    if engine_type:
        creation_message = (
            f"Creating {EngineType.get_label(engine_type)} engine '{engine_name}' with size {engine_size}... "
            f"(this may take several minutes)"
        )
    else:
        creation_message = (
            f"Creating engine '{engine_name}' with size {engine_size}... "
            f"(this may take several minutes)"
        )

    with Spinner(
        creation_message,
        f"Engine '{engine_name}' created!",
        failed_message=None,  # We handle error display ourselves below
    ):
        try:
            provider.create_engine(
                engine_name,
                type=engine_type,
                size=engine_size,
                auto_suspend_mins=auto_suspend_mins,
            )
        except Exception as e:
            # Prefer richer error messages when available
            error_msg = None

            # EngineProvisioningFailed has a format_message() method that provides better error details
            if hasattr(e, "format_message"):
                try:
                    error_msg = getattr(e, "format_message")()
                except Exception:
                    pass

            # Try content/message fallbacks
            if not error_msg and hasattr(e, "content"):
                error_msg = getattr(e, "content", None)
            if not error_msg and hasattr(e, "message"):
                error_msg = str(getattr(e, "message", ""))
            if not error_msg:
                error_msg = str(e)

            raise Exception(error_msg)

#--------------------------------------------------
# Tables
#--------------------------------------------------

def get_color_by_state(state: str) -> str:
    if state and isinstance(state, str):
        state_lower = state.lower()
        if state_lower in ("aborted", "quarantined"):
            return "red"
        elif state_lower == "completed":
            return "white"
        elif state_lower in ("running", "cancelling", "syncing", "pending", "processing"):
            return "bold yellow"
        elif state_lower == "suspended":
            return "dim"
        else:
            return ""
    return ""

def format_value(value) -> str:
    if value is None:
        return "N/A"
    elif isinstance(value, (int, float)):
        return f"{value}"
    elif isinstance(value, list):
        return ", ".join(map(str, value))
    elif isinstance(value, bool):
        return f"{value}"
    elif isinstance(value, datetime):
        return value.strftime("%Y-%m-%d %H:%M:%S")
    elif isinstance(value, timedelta):
        return humanized_duration(int(value.total_seconds() * 1000))
    else:
        return str(value)

def format_row(key: str, value) -> dict:
    result = {}
    result[key] = value
    if "status" in key.lower() or "state" in key.lower():
        result["style"] = get_color_by_state(value)
    if key == "query_size" and isinstance(value, int):
        result[key] = humanized_bytes(value)
    else:
        result[key] = format_value(value)
    return result

def show_dictionary_table(dict, format_fn:Callable|None=None):
    table = Table(show_header=True, border_style="dim", header_style="bold", box=rich_box.SIMPLE_HEAD)
    table.add_column("Field")
    table.add_column("Value")
    for key, value in dict.items():
        if format_fn:
            row = format_fn(key, value)
            table.add_row(key, row.get(key), style=row.get("style"))
        else:
            table.add_row(key, value)
    rich.print(table)

#--------------------------------------------------
# Custom help printer
#--------------------------------------------------


class RichGroup(click.Group):
    def invoke(self, ctx: Context) -> Any:
        """Invoke the CLI command, suppressing tracebacks for handled RAIExceptions.

        Any `RAIException` is expected to already know how to render itself nicely via
        `pprint()`. When such an exception bubbles up to the top-level Click runner,
        Click will otherwise print a full Python traceback, which is noisy for users.
        """
        try:
            return super().invoke(ctx)
        except RAIException as exc:
            # Respect config-based full-trace setting when available.
            try:
                show_full_traces = get_config().get("show_full_traces", SHOW_FULL_TRACES)
            except Exception:
                show_full_traces = SHOW_FULL_TRACES

            if show_full_traces:
                raise

            exc.pprint()
            raise click.exceptions.Exit(1) from None

    def format_help(self, ctx: Context, formatter: HelpFormatter) -> None:
        is_latest, current_ver, latest_ver = is_latest_cli_version()

        global PROFILE
        # @NOTE: I couldn't find a sane way to access the --profile option from here, so insane it is.
        if "--profile" in sys.argv:
            ix = sys.argv.index("--profile") + 1
            if ix < len(sys.argv):
                PROFILE = sys.argv[ix]

        profile = get_config().profile
        if profile == TOP_LEVEL_PROFILE_NAME:
            profile = "[yellow bold]None[/yellow bold]" if not get_config().get("platform", "") else "[ROOT]"

        sio = io.StringIO()
        console = Console(file=sio, force_terminal=True)
        divider(console)
        console.print(f"[bold]Welcome to [green]RelationalAI[/bold] ({current_ver})!")
        console.print()
        if not is_latest:
            console.print(f"[yellow]A new version of RelationalAI is available: {latest_ver}[/yellow] ")
            console.print()
        console.print("rai [magenta]\\[options][/magenta] [cyan]command[/cyan]")

        console.print()
        console.print(f"[magenta]--profile[/magenta][dim] - which config profile to use (current: [/dim][cyan]{profile}[/cyan][dim])")

        unavailable_commands = []
        groups = defaultdict(list)
        for command in self.commands.keys():
            if ":" in command:
                group, _, _ = command.rpartition(":")
                groups[group].append(command)
            else:
                groups[""].append(command)

        console.print()
        for command in groups[""]:
            console.print(f"[cyan]{command}[/cyan][dim] - {self.commands[command].help}")

        for group, commands in groups.items():
            if not group:
                continue

            empty = True
            for command in commands:
                if command_available(self.commands[command]):
                    if empty:
                        empty = False
                        console.print()

                    console.print(f"[cyan]{command}[/cyan][dim] - {self.commands[command].help}")
                else:
                    unavailable_commands.append(command)

        if unavailable_commands:
            platform = get_config().get("platform", "")
            console.print()
            console.print(f"[yellow]Not available on {platform}[/yellow]")
            console.print()
            for command in unavailable_commands:
                console.print(f"[dim yellow]{command} - {self.commands[command].help}")

        divider(console)
        formatter.write(sio.getvalue())
        sio.close()

def filter_profiles_by_platform(config:ConfigFile, platform:str):
    filtered_config = {}
    for profile, props in config.get_combined_profiles().items():
        if profile == TOP_LEVEL_PROFILE_NAME:
            continue
        if props.get("platform") == platform or (
            props.get("platform") is None
            and platform == "azure"
        ):
            filtered_config[profile] = props
    return filtered_config

#--------------------------------------------------
# Imports list
#--------------------------------------------------

def show_imports(imports, showId=False):
    ensure_config()
    if len(imports):
        table = Table(show_header=True, border_style="dim", header_style="bold", box=rich_box.SIMPLE_HEAD)
        cols = {"#": "index"}
        if showId and len(imports) and "id" in imports[0]:
            cols["ID"] = "id"
        if len(imports) and "name" in imports[0]:
            cols["Name"] = "name"
        if len(imports) and "model" in imports[0]:
            cols["Model"] = "model"
        if len(imports) and "created" in imports[0]:
            cols["Created"] = "created"
        if len(imports) and "creator" in imports[0]:
            cols["Creator"] = "creator"
        if len(imports) and "batches" in imports[0]:
            cols["Batches"] = "batches"
        if len(imports) and "status" in imports[0]:
            cols["Status"] = "status"
        if len(imports) and "errors" in imports[0]:
            cols["Errors"] = "errors"

        for label in cols.keys():
            table.add_column(label)

        for index, imp in enumerate(imports):
            imp["index"] = f"{index+1}"
            style = get_color_by_state(imp["status"])
            if imp["created"]:
                imp["created"] = format_value(imp["created"])
            table.add_row(*[imp.get(attr, None) for attr in cols.values()], style=style)
        rich.print(table)
    else:
        rich.print("[yellow]No imports found")

#--------------------------------------------------
# Transactions
#--------------------------------------------------

def show_transactions(transactions, limit):
    if len(transactions):
        table = Table(show_header=True, border_style="dim", header_style="bold", box=rich_box.SIMPLE_HEAD)
        table.add_column("#")
        table.add_column("ID")
        table.add_column("Schema")
        table.add_column("Engine")
        table.add_column("State")
        table.add_column("Created")
        table.add_column("Duration", justify="right")

        added = 0
        for i, txn in enumerate(transactions):
            if added >= limit:
                break

            state = txn.get("state", "")
            duration = txn.get("duration")
            created_on = txn.get("created_on")

            if isinstance(created_on, int):
                created_on = datetime.fromtimestamp(created_on / 1000, tz=pytz.utc)
            if duration is None:
                duration = (datetime.now(created_on.tzinfo) - created_on).total_seconds() * 1000

            table.add_row(
                f"{i+1}",
                txn.get("id"),
                txn.get("database", ""),
                txn.get("engine", ""),
                state,
                created_on.strftime("%Y-%m-%d %H:%M:%S"),
                humanized_duration(int(duration)),
                style=get_color_by_state(state)
            )
            added += 1
        rich.print(table)
    else:
        rich.print("[yellow]No transactions found")

def show_engines(engines):
    if len(engines):
        table = Table(show_header=True, border_style="dim", header_style="bold", box=rich_box.SIMPLE_HEAD)
        table.add_column("#")
        table.add_column("Name")
        # Show type column if present
        table.add_column("Type")
        table.add_column("Size")
        table.add_column("State")
        table.add_column("Created By")
        table.add_column("Created On")
        EngineType = _get_engine_type_api()
        for index, engine in enumerate(engines):
            engine_type = engine.get("type", "")
            type_display = EngineType.get_label_with_value(engine_type) if engine_type and EngineType.is_valid(engine_type) else (engine_type or "")
            created_on = format_value(engine.get("created_on"))
            state_display = format_state_with_color(engine.get("state", ""))
            table.add_row(
                f"{index+1}",
                engine.get("name"),
                type_display,
                engine.get("size"),
                state_display,
                engine.get("created_by", ""),
                created_on,
            )
        rich.print(table)


def show_engine_details(engine: dict[str, Any]) -> None:
    """Print a vertical table of engine details (one field per row)."""
    from relationalai.clients.resources.snowflake import EngineType as _EngineType

    table = Table(
        show_header=True,
        border_style="dim",
        header_style="bold",
        box=rich_box.SIMPLE_HEAD,
    )
    table.add_column("Field")
    table.add_column("Value", overflow="fold")

    engine_type_from_db = engine.get("type", "")
    type_display = (
        _EngineType.get_label_with_value(engine_type_from_db)
        if engine_type_from_db and _EngineType.is_valid(engine_type_from_db)
        else engine_type_from_db
    )

    rows: list[tuple[str, str]] = [
        ("Name", str(engine.get("name", ""))),
        ("Type", str(type_display)),
        ("Size", str(engine.get("size", ""))),
        ("State", str(format_state_with_color(engine.get("state", "")))),
        ("Created By", str(engine.get("created_by", ""))),
        ("Created On", str(format_value(engine.get("created_on")))),
    ]

    # Optional fields (may not exist on older backends).
    for key, label, formatter in [
        ("version", "Version", lambda v: "" if v is None else str(v)),
        ("updated_on", "Updated On", format_value),
        ("suspends_at", "Suspends At", format_value),
    ]:
        if key in engine:
            try:
                val = engine.get(key)
                rows.append((label, str(formatter(val))))
            except Exception:
                rows.append((label, str(engine.get(key))))

    # Auto-suspend minutes is represented differently across backends:
    # - list_engines tends to return auto_suspend_mins
    # - get_engine (Snowflake EngineServiceSQL) returns auto_suspend
    if "auto_suspend_mins" in engine or "auto_suspend" in engine:
        auto_suspend_val = engine.get("auto_suspend_mins", engine.get("auto_suspend"))
        rows.append(("Auto Suspend (mins)", "" if auto_suspend_val is None else str(auto_suspend_val)))

    settings = engine.get("settings")
    if settings in (None, {}, ""):
        settings_str = ""
    elif isinstance(settings, dict):
        settings_str = json.dumps(settings, indent=2, sort_keys=True)
    else:
        settings_str = str(settings)
    rows.append(("Settings", settings_str))

    for field, value in rows:
        table.add_row(field, value)

    rich.print(table)

def exit_with_error(message: str) -> NoReturn:
    rich.print(message, file=sys.stderr)
    exit_with_divider(1)

def exit_with_handled_exception(context: str, exc: Exception) -> NoReturn:
    """Exit with a nicely formatted message for handled RAIExceptions.

    - If `exc` is a RAIException (i.e. produced by an error handler), print its rich
        formatted content via `exc.pprint()`.
    - Otherwise, print a raw one-line error including the context and exception string.
    """
    if isinstance(exc, RAIException):
        exc.pprint()
        sys.exit(1)
    exit_with_error(f"\n\n[yellow]{context}: {exc}")

def exit_with_divider(exit_code: int = 0) -> NoReturn:
    divider()
    sys.exit(exit_code)



from __future__ import annotations
from typing import Optional, Sequence
from relationalai.semantics.internal import internal
from relationalai.semantics.metamodel.ir import (
    Node, Require, Logical, Var, Relation, Lookup, ScalarType, Literal, Exists
)
from relationalai.semantics.metamodel import builtins


"""
Helper functions for converting `Require` nodes with unique constraints to functional
dependencies. The main functionalities provided are:
    1. Check whether a `Require` node is a valid unique constraint representation
    2. Represent the uniqueness constraint as a functional dependency
    3. Check if the functional dependency is structural i.e., can be represented with
       `@function(k)` annotation on a single relation.

=========================== Structure of _legacy_ unique constraints ================================
A `Require` node represents a _unique constraint_ if it meets the following criteria:
 * the `Require` node's `domain` is an empty `Logical` node
 * the `Require` node's `checks` has a single `Check` node
 * the single `Check` node has `Logical` task that is a list of `Lookup` tasks
 * precisely one `Lookup` task in the `Check` uses the `unique` builtin relation name
 * the `unique` lookup has precisely one argument, which is a `TupleArg` or a `tuple`
   containing at least one `Var`
 * all `Lookup` nodes use variables only (no constants)
 * the variables used in the `unique` lookup are a subset of the variables used in other
   lookups
============================================================================================

We use the following _legacy_ unique constraint as the running example.

```
Require
  domain
    Logical
  checks:
    Check
      check:
        Logical
          Person(person::Person)
          first_name(person::Person, firstname::String)
          last_name(person::Person, lastname::String)
          unique((firstname::String, lastname::String))
      error:
        ...
```

=========================== Semantics of unique constraints ================================
A unique constraint states that the columns declared in the `unique` predicate must be
unique in the result of the conjunctive query consisting of all remaining predicates.
============================================================================================

In the running example, the conjunctive query computes a table with 3 columns, the person id
`person::Person`, the first name `firstname::String`, and the last name `lastname::String`.
The uniqueness predicate `unique((firstname::String, lastname::String))` states that no person
can have more than a single combination of first and last name.

The unique constraint in the running example above corresponds to the following functional
dependency.

```
Person(x) ∧ first_name(x, y) ∧ last_name(x, z): {y, z} -> {x}
```

------------------------------ Redundant Type Atoms ----------------------------------------
At the time of writing, PyRel does not yet remove redundant unary atoms. For instance, in
the running example, the atom `Person(person::Person)` is redundant because the type of the
`person` variable is specified in the other two atoms `first_name` and `last_name`.
Consequently, we identify redundant atoms and remove them from the definition of the
corresponding functional dependency.

Formally, a _guard_ atom is any `Lookup` node whose relation name is not `unique`. Now, a
unary guard atom `T(x::T)` is _redundant_ if the uniqueness constraint has a non-unary guard
atom `R(...,x::T,...)`.

================================ Normalized FDs ============================================
Now, the _(normalized)_ functional dependency_ corresponding to a unique constraint is an
object of the form `φ: X → Y`, where :
 1. `φ` is the set of all non-redundant guard atoms.
 2. `X` is the set of variables used in the `unique` atom
 3. `Y` is the set of all other variables used in the constraint
============================================================================================

The normalized functional dependency corresponding to the unique constraints from the running
example is :
```
first_name(person::Person, firstname::String) ∧ last_name(person::Person, lastname::String): {firstname:String, lastname:String} -> {person:Person}
```
Note that the unary atom `Person(person::Person)` is redundant and thus omitted from the
decomposition.

Some simple functional dependencies can, however, be expressed simply with `@function(k)`
attribute of a single relation.  Specifically, a functional dependency `φ: X → Y` is
_structural_ if φ consists of a single atom `R(x1,...,xm,y1,...,yk)` and `X = {x1,...,xm}`.
"""

#
# Checks that an input `Require` node is a valid unique constraint. Returns an empty list if not.
# If yes, we return a list of decompositions of the unique constraint as tuples
# `(all_vars, unique_vars, guard)`, where
# - `all_vars` is the list of all non-existential variables used in the constraint
# - `unique_vars` is the list of variables used in the `unique` atom
# - `guard` is the list of all guard `Lookup` atoms
#
# Two representations are supported:
# 1. Legacy: empty domain, guard atoms and unique atom together in the check's Logical.
# 2. Standard: guard atoms in the domain (possibly wrapped in Exists), unique atom(s) in the
#    check(s). Multiple checks with unique atoms are supported in the standard representation.
#    https://docs.google.com/document/d/1R6FNo5Z3U2UGCQozALIM__Ka5yoLfxq790FyZCnHBmY
# The plan is to transition from the legacy to the standard representation.
#
def _split_unique_require_node(node: Require) -> list[tuple[list[Var], list[Var], list[Lookup]]]:
    legacy = _split_legacy_unique_require_node(node)
    if legacy is not None:
        return [legacy]
    return _split_standard_unique_require_node(node)

#
# Extracts the unique_vars from a `unique` Lookup atom.
# Returns `None` if the atom is not a valid `unique` atom.
#
def _extract_unique_vars(unique_atom: Lookup) -> Optional[list[Var]]:
    if len(unique_atom.args) != 1:
        return None
    if not isinstance(unique_atom.args[0], (internal.TupleArg, tuple)):
        return None
    if len(unique_atom.args[0]) == 0:
        return None
    unique_vars: list[Var] = []
    for arg in unique_atom.args[0]:
        if not isinstance(arg, Var):
            return None
        if arg in unique_vars:
            return None
        unique_vars.append(arg)
    return unique_vars


#
# Legacy representation: empty domain, guard atoms + unique atom in the check.
#
# ```
# Require
#   domain:
#     Logical
#   checks:
#     Check
#       check:
#         Logical
#           R₁(X₁)
#           ...
#           Rₙ(Xₙ)
#           unique((Y))
# ```
#
def _split_legacy_unique_require_node(node: Require) -> Optional[tuple[list[Var], list[Var], list[Lookup]]]:
    if not isinstance(node.domain, Logical):
        return None
    if len(node.domain.body) != 0:
        return None
    if len(node.checks) != 1:
        return None
    check = node.checks[0]
    if not isinstance(check.check, Logical):
        return None

    unique_atom: Optional[Lookup] = None
    guard: list[Lookup] = []
    for task in check.check.body:
        if not isinstance(task, Lookup):
            return None
        if task.relation.name == builtins.unique.name:
            if unique_atom is not None:
                return None
            unique_atom = task
        else:
            guard.append(task)

    if unique_atom is None:
        return None

    # collect variables from guard atoms
    all_vars: list[Var] = []
    for lookup in guard:
        for arg in lookup.args:
            if not isinstance(arg, Var):
                return None
            if arg in all_vars:
                continue
            all_vars.append(arg)

    unique_vars = _extract_unique_vars(unique_atom)
    if unique_vars is None:
        return None

    # check that unique vars are a subset of other vars
    if not set(unique_vars).issubset(set(all_vars)):
        return None

    return list(all_vars), list(unique_vars), guard


#
# Standard representation: guard atoms in the domain (possibly wrapped in Exists),
# unique atom alone in the check.
#
# Without existential variables:
# ```
# Require
#   domain:
#     Logical
#       R₁(X₁)
#       ...
#       Rₙ(Xₙ)
#   checks:
#     Check
#       check:
#         Logical
#           unique((Y))
# ```
#
# With existential variables:
# ```
# Require
#   domain:
#     Logical
#       Exists
#         vars: E
#         task:
#           Logical
#             R₁(X₁)
#             ...
#             Rₙ(Xₙ)
#   checks:
#     Check
#       check:
#         Logical
#           unique((Y))
# ```
#
def _split_standard_unique_require_node(node: Require) -> list[tuple[list[Var], list[Var], list[Lookup]]]:
    if not isinstance(node.domain, Logical):
        return []

    # extract guard atoms and existential vars from domain
    exists_vars: set[Var] = set()
    guard_body: Optional[tuple] = None

    if len(node.domain.body) == 0:
        return []
    elif len(node.domain.body) == 1 and isinstance(node.domain.body[0], Exists):
        # domain is Logical containing a single Exists node
        exists_node = node.domain.body[0]
        exists_vars = set(exists_node.vars)
        if not isinstance(exists_node.task, Logical):
            return []
        guard_body = exists_node.task.body
    else:
        # domain is Logical containing guard atoms directly
        guard_body = node.domain.body

    # extract guard atoms
    guard: list[Lookup] = []
    for task in guard_body:
        if not isinstance(task, Lookup):
            return []
        guard.append(task)

    # A valid unique constraint must have at least one guard atom defining the
    # relation over which uniqueness is enforced.
    if len(guard) == 0:
        return []

    # collect all variables from guard atoms
    all_guard_vars: list[Var] = []
    for lookup in guard:
        for arg in lookup.args:
            if not isinstance(arg, Var):
                return []
            if arg in all_guard_vars:
                continue
            all_guard_vars.append(arg)

    # all_vars = guard vars minus existential vars (preserving order)
    all_vars = [v for v in all_guard_vars if v not in exists_vars]

    # extract unique atom from each check
    if len(node.checks) == 0:
        return []

    results: list[tuple[list[Var], list[Var], list[Lookup]]] = []
    for check in node.checks:
        if not isinstance(check.check, Logical):
            continue
        if len(check.check.body) != 1:
            continue
        unique_lookup = check.check.body[0]
        if not isinstance(unique_lookup, Lookup):
            continue
        if unique_lookup.relation.name != builtins.unique.name:
            continue

        unique_vars = _extract_unique_vars(unique_lookup)
        if unique_vars is None:
            continue

        # check that unique vars are a subset of non-existential vars
        if not set(unique_vars).issubset(set(all_vars)):
            continue

        results.append((all_vars, unique_vars, guard))

    return results


def is_valid_unique_constraint(node: Require) -> bool:
    """
    Checks whether the input `Require` node is a valid unique constraint. See description at
    the top of the file for details.
    """
    return len(_split_unique_require_node(node)) > 0

#
# A unary guard atom `T(x::T)` is redundant if the constraint contains a non-unary atom
# `R(...,x::T,...)`.  We discard all redundant guard atoms in the constructed fd.
#
def _make_fd(all_vars: list[Var], unique_vars: list[Var], guard_atoms: list[Lookup]) -> FunctionalDependency:
    # remove redundant lookups
    redundant_guard_atoms: list[Lookup] = []
    for atom in guard_atoms:
        # the atom is unary A(x::T)
        if len(atom.args) != 1:
            continue
        var = atom.args[0]
        assert isinstance(var, Var)
        # T is a scalar type (which includes entity types)
        var_type = var.type
        if not isinstance(var_type, ScalarType):
            continue
        # the atom is a entity typing T(x::T) i.e., T = A (and hence not a Boolean property)
        var_type_name = var_type.name
        rel_name = atom.relation.name
        if rel_name != var_type_name:
            continue
        # Found an atom of the form T(x::T)
        # check if there is another atom R(...,x::T,...)
        for typed_atom in guard_atoms:
            if len(typed_atom.args) == 1:
                continue
            if var in typed_atom.args:
                redundant_guard_atoms.append(atom)
                break

    guard = [atom for atom in guard_atoms if atom not in redundant_guard_atoms]
    keys = unique_vars
    values = [v for v in all_vars if v not in keys]

    return FunctionalDependency(guard, keys, values)


def normalized_fds(node: Require) -> list[FunctionalDependency]:
    """
    If the input `Require` node is a uniqueness constraint, constructs the list of reduced
    functional dependencies `φ: X -> Y`, where `φ` contains all non-redundant guard atoms,
    `X` are the variables used in the `unique` atom, and `Y` are the remaining variables.
    Returns an empty list if the input node is not a valid uniqueness constraint.
    """
    splits = _split_unique_require_node(node)
    return [_make_fd(all_vars, unique_vars, guard_atoms)
            for all_vars, unique_vars, guard_atoms in splits]

class FunctionalDependency:
    """
    Represents a functional dependency of the form `φ: X -> Y`, where
     - `φ` is a set of `Lookup` atoms
     - `X` and `Y` are disjoint and covering sets of variables used in `φ`
    """
    def __init__(self, guard: Sequence[Lookup], keys: Sequence[Var], values: Sequence[Var]):
        self.guard = tuple(guard)
        self.keys = tuple(keys)
        self.values = tuple(values)
        assert set(self.keys).isdisjoint(set(self.values)), "Keys and values must be disjoint"

        # for structural fd check
        self._is_structural:bool = False
        self._structural_relation:Optional[Relation] = None
        self._structural_rank:Optional[int] = None

        self._determine_is_structural()

        # compute canonical string representation of the fd
        self._canonical_str = self._compute_canonical_str()

    # A functional dependency `φ: X → Y` is _k-functional_ if `φ` consists of a single atom
    # `R(x1,...,xm,y1,...,yk)` and `X = {x1,...,xm}`. Not all functional dependencies are
    # k-functional. For instance, `R(x, y, z): {y, z} → {x}` cannot be expressed with
    # `@function`. neither can `R(x, y) ∧ P(x, z) : {x} → {y, z}`.
    def _determine_is_structural(self):
        if len(self.guard) != 1:
            self._is_structural = False
            return
        atom = next(iter(self.guard))
        atom_vars = atom.args
        if len(atom_vars) <= len(self.keys): # @function(0) provides no information
            self._is_structural = False
            return
        prefix_vars = atom_vars[:len(self.keys)]
        if set(prefix_vars) != set(self.keys):
            self._is_structural = False
            return
        self._is_structural = True
        self._structural_relation = atom.relation
        self._structural_rank = len(atom_vars) - len(self.keys)

    @property
    def is_structural(self) -> bool:
        """
        Whether the functional dependency is functional, i.e., can be represented
        with `@function(k)` annotation on a single relation.
        """
        return self._is_structural

    @property
    def structural_relation(self) -> Relation:
        """
        The structural relation of a functional dependency. Raises ValueError if the functional
        dependency is not structural.
        """
        if not self._is_structural:
            raise ValueError("Functional dependency is not structural")
        assert self._structural_relation is not None
        return self._structural_relation

    @property
    def structural_rank(self) -> int:
        """
        The structural rank k of k-structural fd. Raises ValueError if the structural
        dependency is not k-structural.
        """
        if not self._is_structural:
            raise ValueError("Functional dependency is not structural")
        assert self._structural_rank is not None
        return self._structural_rank

    def __str__(self) -> str:
        guard_str = " ∧ ".join([str(atom) for atom in self.guard]).strip()
        keys_str = ", ".join([str(var) for var in self.keys]).strip()
        values_str = ", ".join([str(var) for var in self.values]).strip()
        return f"{guard_str}: {{{keys_str}}} -> {{{values_str}}}"

    # computes a canonical string representation of the functional dependency
    def _compute_canonical_str(self) -> str:
        # we construct a stable tuple-term representation of the fd
        fd_term = ("fd",)
        for atom in sorted(self.guard, key=lambda x: x.relation.name):
            atom_term = (atom.relation.name,)
            for arg in atom.args:
                if isinstance(arg, Var):
                    atom_term += (("var", arg.name, str(arg.type)),)
                elif isinstance(arg, Literal):
                    atom_term += (("lit", arg.value, str(arg.type)),)
                else:
                    atom_term += (("arg", str(arg)),)
            fd_term += (atom_term,)
        keys_term = tuple(sorted((("var", v.name, str(v.type)) for v in self.keys)))
        values_term = tuple(sorted((("var", v.name, str(v.type)) for v in self.values)))
        fd_term += (("keys", keys_term), ("values", values_term))
        return str(fd_term)

    @property
    def canonical_str(self) -> str:
        """
        A canonical string representation (depends on guard atoms, keys, and values).
        """
        return self._canonical_str

def contains_only_declarable_constraints(node: Node) -> bool:
    """
    Checks whether the input node contains only `Require` nodes annotated with
    `declare_constraint` (or such a node itself).
    """
    # Check if the node itself is a Require node with declarable constraint
    if isinstance(node, Require):
        return is_declarable_constraint(node)

    # Otherwise, check if it is a Logical node containing only declarable constraints
    if not isinstance(node, Logical):
        return False
    if len(node.body) == 0:
        return False
    for task in node.body:
        if not contains_only_declarable_constraints(task):
            return False
    return True

def is_declarable_constraint(node: Node) -> bool:
    """
    Checks whether the input node is a `Require` node annotated with `declare_constraint`.
    """
    if not isinstance(node, Require):
        return False
    return builtins.declare_constraint_annotation in node.annotations



from ..metamodel.compiler import Pass
from ..metamodel.typer import Checker, InferTypes

from ..metamodel.rewrite import (
    DNFUnionSplitter, FormatOutputs, HandleAggregationsAndRanks
)
from .rewrite import (
    AlgorithmPass, AnnotateConstraints, CDC, ConstantsToVars, DeduplicateVars,
    ExtractCommon, ExtractKeys, Flatten, FlattenScript,
    FunctionAnnotations, LiftData, PeriodMath, QuantifyVars, Splinter, SplitMultiCheckRequires,
    UnifyDefinitions,
)

def lqp_passes() -> list[Pass]:
    return [
        Checker(),
        SplitMultiCheckRequires(),
        FunctionAnnotations(),
        AnnotateConstraints(),
        LiftData(),  # Lifts Data nodes to top-level GNF predicates (before Splinter for null safety).
        CDC(), # specialize to physical relations before extracting nested and typing
        InferTypes(),

        ## Start restructuring the metamodel to resemble LQP structure.

        ## Handle GNF and outer join semantics
        DNFUnionSplitter(), # Handle unions that require DNF decomposition
        ExtractKeys(), # Create a logical for each valid combinations of keys
        FormatOutputs(),

        ## Move nested tasks to the top level, and ensure the root logical has the expected
        ## structure.
        ExtractCommon(), # Extracts tasks that will become common after Flatten into their own definition
        Flatten(), # Move nested tasks to the top level, and various related things touched along the way
        FlattenScript(), # Additional flattening specific to scripts
        HandleAggregationsAndRanks(), # Handle aggregation and rank dependencies

        ## Post-flattening passes
        # Period Math breaks relational assumptions used by dependency analysis so we place
        # it after work relying on that.
        PeriodMath(),  # Rewrite date period uses.
        Splinter(), # Splits multi-headed rules into multiple rules

        ## Final cleanup passes to match with LQP needs.
        ConstantsToVars(),  # Turns constants in Updates and Outputs into vars.
        DeduplicateVars(),  # Deduplicates vars in Updates and Outputs.
        QuantifyVars(), # Adds missing existentials
        AlgorithmPass(),  # Passes to make sure we are following loopy invariants
        UnifyDefinitions(), # Unify relations with multiple definitions.
    ]



from __future__ import annotations

from pandas import DataFrame
from typing import Any, Union, Tuple, Literal, TYPE_CHECKING

from relationalai.clients.config import Config
from relationalai.semantics.metamodel import Model, Task, ir
from relationalai.semantics.metamodel.visitor import collect_by_type
if TYPE_CHECKING:
    from relationalai.semantics.internal.internal import Model as InternalModel

from .util import NameCache

import rich

# global flag to suppress type errors from being printed
SUPPRESS_TYPE_ERRORS = False

class Executor():
    """ Interface for an object that can execute the program specified by a model. """
    def execute(self, model: Model, task:Task, format:Literal["pandas", "snowpark"]="pandas",
                export_to: Any = None, update: bool = False, meta: dict[str, Any] | None = None,
                source_info: dict[str, Any] | None = None) -> Union[DataFrame, Any]:
        raise NotImplementedError(f"execute: {self}")

    def _compute_cols(self, task: ir.Task, final_model: ir.Model|None) -> Tuple[list[str], list[str], list[int]]:
        cols = [] # all cols in output
        extra_cols = [] # all key cols not in output
        key_locs = [] # locations of output key cols in all output cols

        # we assume only queries have outputs
        original_outputs = collect_by_type(ir.Output, task) if task else None
        outputs = collect_by_type(ir.Output, final_model) if final_model else None

        # there are some outputs, and some have keys
        if original_outputs and outputs and not all(not out.keys for out in outputs):
            assert len(original_outputs) == 1
            original_output = original_outputs[0]
            original_cols = []
            original_cols_val = []

            keys = outputs[0].keys
            assert keys

            for out in outputs:
                assert out.keys is not None
                assert set(out.keys) == set(keys), "outputs with different key sets in the same query"

            for (idx, (alias, val)) in enumerate(original_output.aliases):
                if not alias:
                    continue
                original_cols.append(alias)
                original_cols_val.append(val)
                if isinstance(val, ir.Var) and val in keys:
                    key_locs.append(idx)

            extra_cols = []
            name_cache = NameCache(start_from_one=True)
            for key in keys:
                if isinstance(key, ir.Var) and key not in original_cols_val:
                    extra_cols.append(f"hidden_{name_cache.get_name(key.id, key.name)}")
            cols = original_cols + extra_cols
        elif outputs:
            cols = [alias for alias, _ in outputs[-1].aliases if alias]

        return cols, extra_cols, key_locs

    def _postprocess_df(self, config: Config, df: DataFrame, extra_cols: list[str]) -> DataFrame:
        if bool(config.get("compiler.debug_hidden_keys", False)):
            rich.print("[blue]DataFrame with extra hidden columns:[/blue]")
            rich.print(f"[black]{df}[/black]")
        if not df.empty:
            for col in extra_cols:
                if col in df.columns:
                    df = df.drop(col, axis=1)
        return df

    def export_to_csv(self, model: "InternalModel", query) -> str:
        ### Only implemented in the LQP executor for now.
        raise NotImplementedError(f"export_to_csv is not supported by {type(self).__name__}")

    def batch_export_to_csv(self, model: "InternalModel", queries: list) -> list[str]:
        ### Only implemented in the LQP executor for now.
        raise NotImplementedError(f"batch_export_to_csv is not supported by {type(self).__name__}")



"""
Transaction progress updater that manages displaying execution tree as subtasks.

This module handles displaying processed transaction progress data,
keeping display concerns separate from data processing.
"""
from typing import Any, Dict, List, Optional

from relationalai.tools.cli_controls import TaskProgress, NotebookTaskProgress
from relationalai.tools.txn_progress_tree import (
    ProcessedTask,
    ProgressSummaryItem,
    build_progress_tree,
    get_ivm_relative_update_size,
    get_loop_iteration,
)
from relationalai.util.format import format_duration


def should_include_task(task: ProcessedTask) -> bool:
    """Filter out error tasks that aren't useful for display."""
    return not task.task_name.startswith(":error_pyrel_error_attrs")


def format_task_tree(task: ProcessedTask, base_indent: str = "") -> str:
    """
    Format a task tree as a multi-line string for display.
    
    This is a pure function for easy testing. The base_indent parameter
    accounts for the progress system's prefix (5 spaces = "   ➜ ").
    
    Args:
        task: Root task to format
        base_indent: Base indentation (empty for first line, "     " for subsequent)
        
    Returns:
        Multi-line string representation of the task tree
    """
    lines = _format_task_tree_lines(task, indent_level=0, base_indent=base_indent)
    return "\n".join(lines)


def _format_task_tree_lines(
    task: ProcessedTask,
    indent_level: int,
    base_indent: str
) -> List[str]:
    """
    Recursively build lines for a task tree with proper indentation.
    
    Args:
        task: Task node to format
        indent_level: Current indentation level (0 for root)
        base_indent: Base indentation for the first line of this task
        
    Returns:
        List of formatted lines for this task and its children
    """
    lines = []
    
    # Calculate indentation for this line
    # Each indent level adds 2 spaces
    task_indent = base_indent + ("  " * indent_level)
    
    # Format and add this task
    description = _format_task_description(task)
    lines.append(task_indent + description)
    
    # After the first line, all subsequent lines need full base_indent
    # The progress system adds "   ➜ " (5 chars) before the first line only
    child_base_indent = "     " if not base_indent else base_indent
    
    # Add warnings as children
    for warning_code, message in task.warnings:
        warning_indent = child_base_indent + ("  " * (indent_level + 1))
        lines.append(f"{warning_indent}⚠️  {warning_code}: {message}")
    
    ivm_relative_update_size = get_ivm_relative_update_size(task)
    if ivm_relative_update_size:
        indent = child_base_indent + ("  " * (indent_level + 1))
        lines.append(f"{indent}Updating relation incrementally. Updated rows: {ivm_relative_update_size:.1f}%")

    loop_iteration = get_loop_iteration(task)
    if loop_iteration:
        if task.end_time:
            return [f"Loop {task.task_name}: terminated after {loop_iteration} iterations"]
        else:
            return [f"Loop {task.task_name} running on iteration {loop_iteration}"]

    # Add children
    for child in task.children:
        if should_include_task(child):
            child_lines = _format_task_tree_lines(child, indent_level + 1, child_base_indent)
            lines.extend(child_lines)
    
    return lines


def _format_task_description(task: ProcessedTask) -> str:
    """Format a single task description for display."""
    name = task.task_name

    has_warnings = len(task.warnings) > 0
    if has_warnings:
        warning_count = len(task.warnings)
        warning_text = f" ⚠️  {warning_count} warning" + ("s" if warning_count > 1 else "")
    else:
        warning_text = ""
    
    if task.duration is not None:
        duration_str = format_duration(task.duration)
        return f"{name} ({duration_str}){warning_text}"
    else:
        return f"{name} (Evaluating...){warning_text}"


class TxnProgressUpdater:
    """Manages displaying transaction execution tree as subtasks in a TaskProgress."""
    
    def __init__(self, progress: TaskProgress | NotebookTaskProgress | Any, enabled: bool = False):
        """
        Initialize the updater.
        
        Args:
            progress: The TaskProgress or NotebookTaskProgress instance to update
            enabled: Whether to display internal transaction progress
        """
        self.progress = progress
        self.enabled = enabled
    
    def update(self, internal_txn_progress: Optional[Dict]) -> None:
        """
        Update the progress display with transaction execution tree.
        
        Args:
            internal_txn_progress: The internal transaction progress dict with 'tasks' key
        """
        if not self.enabled or internal_txn_progress is None:
            return
        
        if "tasks" not in internal_txn_progress:
            return
        
        # Process data into immutable tree structure
        task_tree = build_progress_tree(internal_txn_progress["tasks"])
        
        # Display the processed tasks
        self._display_tasks(task_tree)
    
    def _display_tasks(self, task_tree: List[ProcessedTask]) -> None:
        """
        Display task tree by updating the progress display.
        
        Only root tasks are added as sub-tasks. Each root's description
        includes its entire subtree rendered as a multi-line string with
        proper indentation that accounts for the progress system's prefix.
        
        Args:
            task_tree: Immutable tree of processed tasks to display
        """
        current_root_ids = set()
        
        # Process each root task
        for root in task_tree:
            if not should_include_task(root):
                continue
                
            current_root_ids.add(root.task_id)
            
            # Build the full multi-line description for this root and its children
            description = format_task_tree(root, base_indent="")
            
            # Get or create the progress task for this root, using the transaction task_id
            if root.task_id in self.progress.list_sub_tasks():
                self.progress.update_sub_task(root.task_id, description)
            else:
                self.progress.add_sub_task(description, task_id=root.task_id)
            
            # Mark as complete if the root task is finished
            if root.duration is not None:
                completed_tasks = self.progress.get_completed_tasks()
                if root.task_id not in completed_tasks:
                    self.progress.complete_sub_task(root.task_id, record_time=False)
        
        # Remove root tasks that are no longer roots. This happens when a child task's parent appears.
        tasks_to_remove = set(self.progress.list_sub_tasks()) - current_root_ids
        for task_id in tasks_to_remove:
            self.progress.remove_sub_task(task_id, animate=False)


def format_progress_summary(
    summary_items: List[ProgressSummaryItem],
    total_duration: float
) -> str:
    """
    Format progress summary items for display.
    
    Args:
        summary_items: List of summary items to format
        total_duration: Total transaction duration in seconds
        
    Returns:
        Formatted multi-line string showing task names, durations, and percentages
    """
    if not summary_items or total_duration <= 0:
        return ""
    
    sorted_items = sorted(summary_items, key=lambda item: item.wall_clock_duration, reverse=True)

    lines = []
    for item in sorted_items:
        if item.loop_iteration:
            lines.append(f"  Loop {item.task_name}: terminated after {item.loop_iteration} iterations")
        else:
            duration_str = format_duration(item.wall_clock_duration)
            percentage = (item.wall_clock_duration / total_duration) * 100
            percentage_str = f"{percentage:.1f}%"
            lines.append(f"  {item.task_name}: {duration_str} ({percentage_str})")
            if item.ivm_relative_update_size:
                lines.append(f"    Relation incrementally updated. Updated rows: {item.ivm_relative_update_size:.1f}%")
    
    return "\n".join(lines)



from __future__ import annotations

import time
from typing import Dict, Optional, TYPE_CHECKING

from relationalai import debugging
from relationalai.clients.util import poll_with_specified_overhead
from relationalai.clients.config import Config
from relationalai.tools.cli_controls import create_progress
from relationalai.tools.txn_progress_tree import build_progress_tree, compute_progress_summary
from relationalai.tools.txn_progress_display import format_progress_summary
from relationalai.util.format import format_duration
from relationalai.tools.constants import USE_DIRECT_ACCESS
from relationalai.tools.txn_progress_display import TxnProgressUpdater

if TYPE_CHECKING:
    from relationalai.clients.resources.snowflake import Resources
    from relationalai.clients.resources.snowflake.snowflake import TxnStatusResponse

# Polling behavior constants
POLL_OVERHEAD_RATE = 0.1  # Overhead rate for exponential backoff
MAX_POLL_DELAY_SECONDS = 20
MAX_POLL_DELAY_FOR_DIRECT_ACCESS_SECONDS = 10

# Text color constants
GREEN_COLOR = '\033[92m'
GRAY_COLOR = '\033[90m'
ENDC = '\033[0m'

PRINT_TXN_PROGRESS_FLAG = "print_txn_progress"
PRINT_INTERNAL_TXN_PROGRESS_FLAG = "print_txn_progress_internal"

def should_print_txn_progress(config: Config) -> bool:
    return bool(config.get(PRINT_TXN_PROGRESS_FLAG, True))

def should_print_internal_txn_progress(config) -> bool:
    return bool(config.get(PRINT_INTERNAL_TXN_PROGRESS_FLAG, False))

class ExecTxnPoller:
    """
    Encapsulates the polling logic for exec_async transaction completion.
    """

    def __init__(
        self,
        config: Config,
        engine: str,
        model: str,
        resource: Optional["Resources"] = None,
        txn_id: Optional[str] = None,
        headers: Optional[Dict] = None,
        txn_start_time: Optional[float] = None,
        source_info: Optional[Dict] = None,
        suppress_creation_errors: bool = False,
    ):
        self.print_txn_progress = should_print_txn_progress(config)
        self.print_internal_txn_progress = should_print_internal_txn_progress(config)
        self.res = resource
        self.txn_id = txn_id
        self.headers = headers or {}
        self.engine = engine
        self.model = model
        self.source_info = source_info or {}
        self.txn_start_time = txn_start_time or time.time()
        self.last_status: Optional[TxnStatusResponse] = None
        self.max_poll_delay = MAX_POLL_DELAY_SECONDS
        if config.get("use_direct_access", USE_DIRECT_ACCESS):
            self.max_poll_delay = MAX_POLL_DELAY_FOR_DIRECT_ACCESS_SECONDS
        self.txn_progress_updater: Optional[TxnProgressUpdater] = None
        self.suppress_creation_errors = suppress_creation_errors
        self.evaluation_progress = {
            'completed_relations': 0,
        }

    def __enter__(self) -> ExecTxnPoller:
        if not self.print_txn_progress:
            return self
        print(self._format_source_label())

        self.progress = create_progress(
            description=lambda: self.description_with_timing(),
            success_message="",  # We'll handle this ourselves
            leading_newline=False,
            trailing_newline=False,
            show_duration_summary=False,
        )
        self.progress.__enter__()
        # Initialize the progress updater to manage execution tree display
        self.txn_progress_updater = TxnProgressUpdater(
            self.progress,
            enabled=self.print_internal_txn_progress
        )
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        if not self.print_txn_progress:
            return

        if exc_type is not None and self.txn_id is None and self.suppress_creation_errors:
            # Transaction creation failed before we got a txn_id, and the
            # caller indicated it will handle retries (e.g., _exec_with_gi_retry
            # retrying after "database not found" on first model access).
            # Silently dismiss the progress display — the retry path will
            # create a new progress on the next attempt.
            live = getattr(self.progress, 'live', None)
            if live is not None:
                live.transient = True
                live.stop()
                self.progress.live = None  # type: ignore[union-attr]
            # Prevent failure message output in both terminal and notebook
            # progress displays (TaskProgress checks self.live which is now
            # None; NotebookTaskProgress checks self.failure_message).
            self.progress.failure_message = None  # type: ignore[assignment]
            self.progress.__exit__(exc_type, exc_value, traceback)
            return

        # Update to success message with duration
        total_duration = time.time() - self.txn_start_time
        txn_id = self.txn_id
        summary = query_progress_message(self.evaluation_progress, True, total_duration, txn_id)
        if self.print_internal_txn_progress and self.last_status and self.last_status.progress:
            task_tree = build_progress_tree(self.last_status.progress["tasks"])
            summary += "\n" + format_progress_summary(compute_progress_summary(task_tree), total_duration)
        self.progress.update_main_status(summary)
        self.progress.__exit__(exc_type, exc_value, traceback)
        return

    def _get_last_progress(self) -> Optional[Dict]:
        """Get last transaction progress message if enabled and available."""
        if self.print_txn_progress and self.last_status:
            return self.last_status.progress
        return None

    def poll(self) -> bool:
        """
        Poll for transaction completion with interactive progress display.

        Returns:
            True if transaction completed successfully, False otherwise
        """
        if not self.txn_id:
            raise ValueError("Transaction ID must be provided for polling.")
        else:
            txn_id = self.txn_id

        if self.print_txn_progress:
            # Update the main status to include the new txn_id
            self.progress.update_main_status_fn(
                lambda: self.description_with_timing(txn_id),
            )

        # Don't show duration summary - we handle our own completion message
        def check_status() -> bool:
            """Check if transaction is complete."""
            if self.res is None:
                raise ValueError("Resource must be provided for polling.")
            self.last_status = self.res._check_exec_async_status(txn_id, headers=self.headers)
            # Update the execution tree display if enabled
            if self.txn_progress_updater:
                progress_data = self.last_status.progress if self.last_status else None
                self.txn_progress_updater.update(progress_data)
            self.update_evaluation_progress()
            return self.last_status.finished

        with debugging.span("wait", txn_id=txn_id):
            poll_with_specified_overhead(check_status, overhead_rate=POLL_OVERHEAD_RATE, max_delay=self.max_poll_delay)

        return True

    def _format_source_label(self) -> str:
        """Format source location as 'Query @ file:line' or 'Query' if no source info."""
        file = self.source_info.get("file")
        line = self.source_info.get("line")
        if file:
            import os
            basename = os.path.basename(file)
            return f"Query @ {basename}:{line}"
        else:
            return "Querying..."

    def description_with_timing(self, txn_id: str | None = None) -> str:
        elapsed = time.time() - self.txn_start_time
        return query_progress_message(self.evaluation_progress, False, elapsed, txn_id)

    def update_evaluation_progress(self):
        """Update self.evaluation_progress from the txn_progress returned by check exec status."""
        txn_progress = self._get_last_progress()
        if txn_progress is None:
            return (0,0)
        completed_relations = 0
        in_progress_relations = 0
        for task_key, task in txn_progress.get('tasks', dict()).items():
            # print(task_key, task)
            if 'end_time' in task:
                completed_relations += 1
            else:
                in_progress_relations += 1
        self.evaluation_progress['completed_relations'] = completed_relations
        self.evaluation_progress['in_progress_relations'] = in_progress_relations

def relations_progress_body(in_progress: int, completed: int) -> str:
    return (
        # TODO(PR): Decide on verbiage. Relations? Views?
        f"   Relations Materialized: {completed} completed" +
        (f" ({in_progress} in progress)" if in_progress > 0 else "")
    )

def query_progress_message(evaluation_progress: dict, completed: bool, duration: float, id: str | None = None) -> str:
    relations_completed = evaluation_progress.get('completed_relations', 0)
    relations_in_progress = evaluation_progress.get('in_progress_relations', 0)
    if completed:
        # Assume all in-progress are completed. (This can happen if the engine doesn't send
        # its final observabilty updates before we mark the txn completed.)
        relations_completed += relations_in_progress
        relations_in_progress = 0
        # # There will always be at least the select relation, but it usually finishes before
        # # the last observability update, so we artifically add that.
        # relations_completed = max(1, relations_completed)

    # Don't print sub-second decimals, because it updates too fast and is distracting.
    duration_str = format_duration(duration)
    if id is None:
        # Print with whitespace to align with the end of the transaction ID
        return f"Submitting job...                            {duration_str}"
    else:
        out = f"ID: {GRAY_COLOR}{id}{ENDC}  {duration_str}"

    if relations_completed + relations_in_progress > 0:
        out += f"\n{ENDC}  |  Evaluating relations: {relations_completed} completed"
        if relations_in_progress > 0:
            out += f" ({relations_in_progress} in progress)"
        out += ENDC

    return out



from __future__ import annotations
from dataclasses import dataclass
from typing import Optional, TypeVar
from typing import Tuple

from relationalai.semantics.metamodel import builtins, ir, factory as f, helpers
from relationalai.semantics.metamodel.compiler import Pass, group_tasks
from relationalai.semantics.metamodel.util import NameCache, OrderedSet, ordered_set
from relationalai.semantics.metamodel import dependency
from relationalai.semantics.metamodel.typer.typer import to_type

from ..builtins import lifted_data

class Flatten(Pass):
    """
    Traverses the model's root to flatten it as much as possible. The result of this pass is
    a Logical root where all nested tasks that represent a rule in Rel are extracted to the
    top level. Additionally, any Sequence is promoted to the top level Logical (but
    encapsulated by a Logical).

    - nested logical with updates becomes a top-level logical (a rule)

    From:
        Logical
            Logical
                lookup1   <- scope is spread
                Logical
                    lookup2
                    derive foo
                Logical
                    lookup3
                    derive bar
    To:
        Logical
            Logical
                lookup1
                lookup2
                derive foo
            Logical
                lookup1
                lookup3
                derive bar

    - a match becomes a top-level logical for each branch, each writing into its own temporary
    relation and a lookup from the last relation. The top-level logical for a branch derives
    into the temporary relation negating the previous branch:

    From:
        Logical
            Logical
                Match
                    Logical
                        lookup1
                    Logical
                        lookup2
                output
    To:
        Logical
            Logical
                lookup1
                derive tmp1
            Logical
                Union            <- tmp1() or (not temp1() and lookup2())
                    lookup tmp1
                    Logical
                        Not
                            lookup tmp1
                        lookup2
                        derive tmp2
            Logical
                lookup tmp2
                output

    - a Sequence is promoted to the top level Logical, encapsulated by a Logical:
    From:
        Logical
            Logical
                lookup
                derive foo
            Sequence
                Logical
                    ...
                Loop
                    Sequence
                        ...
                Logical
                    ...
    To:
        Logical
            Logical
                lookup
                derive foo
            Logical
                Sequence
                    Logical
                        ...
                    Loop
                        Sequence
                            ...
                    Logical
                        ...
    """

    def __init__(self):
        super().__init__()
        self.name_cache = NameCache(start_from_one=True)

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        # create the context
        ctx = Flatten.Context(model, options)

        # rewrite the root
        result = self.handle(model.root, ctx)

        # the new body contains the extracted top level logicals and maybe the rewritten root
        body = ctx.rewrite_ctx.top_level if result.replacement is None else ctx.rewrite_ctx.top_level + [result.replacement]

        # create the new model, updating relations and root
        return ir.Model(
            model.engines,
            OrderedSet.from_iterable(model.relations).update(ctx.rewrite_ctx.relations).frozen(),
            model.types,
            ir.Logical(model.root.engine, tuple(), tuple(body))
        )

    #--------------------------------------------------
    # IR handlers
    #--------------------------------------------------

    class Context():
        def __init__(self, model: ir.Model, options: dict):
            self.rewrite_ctx = helpers.RewriteContext()
            self.info = dependency.analyze_constraints(model.root)
            self.options = options
            # when we extract a nested logical (due to match, etc), we also
            # extract together its dependencies, and we leave behind a different nested
            # logical as a reference, with a lookup to the extracted relation. This dict
            # keeps a map from the reference logical id to the set of tasks that where
            # extracted for it. This is used to optimize the resulting code, since we can
            # assume that the reference logical already includes those tasks.
            self.included: dict[int, OrderedSet[ir.Task]] = dict()
            # These are extra tasks to be added to the body of a logical when it is
            # extracted as a top-level logical. For example, when handling a match, we add
            # a couple of tasks here in case the branches have updates or aggregations that
            # cause the branch to be extracted, so that they include these tasks.
            self.extra_tasks = []


    @dataclass
    class HandleResult():
        replacement: Optional[ir.Task]

    def handle(self, task: ir.Task, ctx: Context) -> Flatten.HandleResult:
        if isinstance(task, ir.Logical):
            return self.handle_logical(task, ctx)
        elif isinstance(task, ir.Match):
            return self.handle_match(task, ctx)
        elif isinstance(task, ir.Require):
            return self.handle_require(task, ctx)
        elif isinstance(task, ir.Sequence):
            return self.handle_sequence(task, ctx)
        else:
            return Flatten.HandleResult(task)


    def handle_logical(self, task: ir.Logical, ctx: Context):
        # keep track of what's the result of handling nested composites
        composites = group_tasks(task.body, {
            "composites": helpers.COMPOSITES
        })["composites"]
        all_composites_removed = len(composites) > 0

        if any(anno.relation is lifted_data for anno in task.annotations):
            # if this is a lifted data logical, we can skip handling of the body and
            # flatten it directly.
            ctx.rewrite_ctx.top_level.append(task)
            return Flatten.HandleResult(None)

        # recursively handle children, collecting the replacements in the body
        body:OrderedSet[ir.Task] = ordered_set()
        for child in task.body:
            result = self.handle(child, ctx)
            if result.replacement is not None:
                ctx.info.replaced(child, result.replacement)
                extend_body(body, result.replacement)
                # nested composite was not completely removed
                if child in composites:
                    all_composites_removed = False

        # Filter out empty logicals from the body
        new_body:OrderedSet[ir.Task] = ordered_set()
        for b in body:
            if isinstance(b, ir.Logical):
                if not b.body:
                    # empty logical, skip
                    continue
            new_body.add(b)

        body = new_body

        # all children were extracted or all composites were removed without any effects
        # left and no outputs (so no way for outer dependencies), drop this logical
        if not body or (all_composites_removed and not any([isinstance(t, helpers.EFFECTS) for t in body]) and not ctx.info.task_outputs(task)):
            return Flatten.HandleResult(None)

        # now process the rewritten body
        groups = group_tasks(body.get_list(), {
            "outputs": ir.Output,
            "updates": ir.Update,
        })

        # If there are outputs, flatten each into its own top-level rule, along with its
        # dependencies.
        if groups["outputs"]:
            # Analyze the dependencies in the newly rewritten body
            new_logical = ir.Logical(task.engine, task.hoisted, tuple(body))
            info = dependency.analyze(new_logical)

            for output in groups["outputs"]:
                assert(isinstance(output, ir.Output))
                deps = info.task_dependencies(output)
                new_body = OrderedSet.from_iterable(t.clone() for t in deps)
                new_body.update(t.clone() for t in ctx.extra_tasks)
                new_body.add(output.clone())
                ctx.rewrite_ctx.top_level.append(ir.Logical(task.engine, task.hoisted, tuple(new_body), task.annotations))

            return Flatten.HandleResult(None)

        # if there are updates, extract as a new top level rule
        if groups["updates"]:
            # add task dependencies to the body
            body.prefix(t.clone() for t in ctx.info.task_dependencies(task))
            # potentially add context extra tasks
            body.update(t.clone() for t in ctx.extra_tasks)
            ctx.rewrite_ctx.top_level.append(ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations))
            return Flatten.HandleResult(None)

        return Flatten.HandleResult(ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations))


    def handle_match(self, match: ir.Match, ctx: Context):
        # TODO: how to deal with malformed input like this?
        if not match.tasks:
            return Flatten.HandleResult(match)

        body = ctx.info.task_dependencies(match)
        exposed_vars = self.compute_exposed_vars(match, match.tasks, ctx)

        # negation len is the number of wildcards when negating a reference
        outputs = ctx.info.task_outputs(match)
        negation_len = len(outputs) if outputs else 0

        # collect the references to branch rules so that we union in the end
        references = []

        # a negated reference to the previous branch (e.g. "not branch1(...)")
        negated_reference = None

        # number of negated references currently in context (e.g. in the 3rd branch there
        # will be 2 negated references, "not branch1(...) and not branch2(...)")
        negated_references = 0
        for branch in match.tasks:
            # if there's a previous connection, add the negation of it to extra tasks
            if negated_reference:
                ctx.extra_tasks.append(negated_reference)
                negated_references += 1

            # the connection relation to be used for this branch
            name = helpers.create_task_name(self.name_cache, branch, "_match")
            relation = helpers.create_connection_relation(branch, exposed_vars, ctx.rewrite_ctx, name)
            branch_update = f.derive(relation, exposed_vars)
            ctx.extra_tasks.append(branch_update)

            # handle the branch
            result = self.handle(branch, ctx)

            # if there's no replacement, the branch was already extracted (e.g. by an update),
            # otherwise, we need to extract the branch as a rule writing into the connection
            if result.replacement:
                ctx.info.replaced(branch, result.replacement)

                branch_body: OrderedSet[ir.Task] = OrderedSet.from_iterable(body)
                extend_body(branch_body, result.replacement)
                # potentially add context extra tasks (negations of previous branches + derive into this branch)
                branch_body.update(ctx.extra_tasks)
                # extract the body
                ctx.rewrite_ctx.top_level.append(ir.Logical(match.engine, tuple(), tuple(branch_body)))

            # cleanup the branch update from context (but keep the negations)
            ctx.extra_tasks.pop()

            # record a reference to the branch connection and adjust the negated_reference
            reference = f.lookup(relation, exposed_vars)
            negated_reference = negate(reference, negation_len)
            references.append(reference)

        # we accumulate negations, in the end we remove them all
        for x in range(negated_references):
            ctx.extra_tasks.pop()

        # return a union of the references to the branches
        return Flatten.HandleResult(f.union(references, match.hoisted))

    def compute_exposed_vars(self, task: ir.Task, branches: Tuple[ir.Task, ...], ctx: Context) -> list:
        """ Helper to compute the set of exposed vars for handle_match and handle_union. """
        exposed_vars = set_union(ctx.info.task_inputs(task), ctx.info.task_outputs(task))
        if exposed_vars:
            return exposed_vars
        # no exposed vars, so we try to use the common vars across all branches as context
        common_vars = helpers.collect_vars(branches[0])
        for branch in branches:
            branch_vars = helpers.collect_vars(branch)
            for v in common_vars:
                if v not in branch_vars:
                    common_vars.remove(v)
        # TODO - we should raise an error here if the set is empty
        return common_vars.get_list()


    def handle_require(self, req: ir.Require, ctx: Context):
        # only extract the domain if it is a somewhat complex Logical and there's more than
        # one check, otherwise insert it straight into all checks
        if builtins.discharged_annotation in req.annotations:
            # remove discharged Requires
            return Flatten.HandleResult(None)
        elif builtins.declare_constraint_annotation in req.annotations:
            # leave Requires that are declared constraints
            return Flatten.HandleResult(req)
        else:
            # generate logic for remaining requires
            domain = req.domain
            if len(req.checks) > 1 and isinstance(domain, ir.Logical) and len(domain.body) > 1:
                body = OrderedSet.from_iterable(domain.body)
                vars = helpers.hoisted_vars(domain.hoisted)
                name = helpers.create_task_name(self.name_cache, req)
                connection = helpers.extract(req, body, vars, ctx.rewrite_ctx, name)
                domain = f.logical([f.lookup(connection, vars)], vars)

            for check in req.checks:
                # only generate logic for checks that have errors and not discharged
                if check.error and builtins.discharged_annotation not in check.annotations:
                    handled_check_result = self.handle(check.check, ctx)
                    if handled_check_result.replacement:
                        body = ordered_set()
                        body.add(domain)
                        body.add(ir.Not(req.engine, handled_check_result.replacement))
                        if (isinstance(check.error, ir.Logical)):
                            body.update(check.error.body)
                        else:
                            # this is more general but may trip the current splinter
                            body.add(check.error)
                        ctx.rewrite_ctx.top_level.append(ir.Logical(req.engine, tuple(), tuple(body)))

        # currently we just drop the Require, but we should keep it here and link the
        # extracted logicals to it
        return Flatten.HandleResult(None)

    def handle_sequence(self, task: ir.Sequence, ctx: Context):
        new_logical = f.logical(
            body = [task],
            engine = task.engine
        )
        ctx.rewrite_ctx.top_level.append(new_logical)
        return Flatten.HandleResult(None)

#--------------------------------------------------
# Helpers
#--------------------------------------------------

T = TypeVar('T')
def set_union(s1: Optional[OrderedSet[T]], s2: Optional[OrderedSet[T]]) -> list:
    """ Return the union of s1 | s2, accounting for when any is None. """
    if s1 and s2:
        return (s1 | s2).get_list()
    if s1:
        return s1.get_list()
    if s2:
        return s2.get_list()
    return []

def negate(lookup: ir.Lookup, values: int):
    """
    Return a negation of this reference, where the last `values` arguments are to
    be replaced by wildcards (i.e. len(reference.args) - values are keys so they need
    to be bound in the Not.)
    """
    args = []
    i = 0
    last = len(lookup.args) - values
    for arg in lookup.args:
        args.append(f.wild(to_type(arg))) if i >= last else args.append(arg)
        i += 1

    return ir.Not(lookup.engine, f.lookup(lookup.relation, args))

def merge_var_list(vars: list[ir.Var], hoisted: tuple[ir.VarOrDefault, ...]) -> list[ir.VarOrDefault]:
    """ Merge vars and hoisted, making sure that hoisted vars have precedence since they may have defaults. """
    r = []
    hoisted_vars = helpers.hoisted_vars(hoisted)
    for v in vars:
        if v not in hoisted_vars:
            r.append(v)
    r.extend(hoisted)
    return r

def extend_body(body: OrderedSet[ir.Task], extra: ir.Task):
    """ Add the extra task to the body, but if the extra is a simple logical, just
    inline its subtasks. """
    if isinstance(extra, ir.Logical):
        if extra.hoisted:
            # hoists, remove things that are already in the body to avoid duplicates
            logical_body = []
            for t in extra.body:
                if t not in body:
                    logical_body.append(t)
            if len(logical_body) == len(extra.body):
                # no duplicates
                body.add(extra)
            else:
                # some duplicate, remove them
                body.add(ir.Logical(
                    extra.engine,
                    extra.hoisted,
                    tuple(logical_body)
                ))
        else:
            body.add(extra)
    else:
        body.add(extra)



#----------------------------------------------------------------------------------------------
# This is a custom LQP rewriter that filters extra columns from CSV export. It is used in the
# LQP executor, when the format="csv", to ensure only intended columns are being exported.
#----------------------------------------------------------------------------------------------

from dataclasses import replace
from relationalai.semantics.lqp import lqp_bundle

class PartitioningDisabler:
    """Disables output file partitioning for CSV exports by setting partition_size=-1."""

    def rewrite_read(self, read: lqp_bundle.ir.Read) -> lqp_bundle.ir.Read:
        if not isinstance(read.read_type, lqp_bundle.ir.Export):
            return read
        config = read.read_type.config
        new_config = replace(config, partition_size=-1)
        new_read_type = replace(read.read_type, config=new_config)
        return replace(read, read_type=new_read_type)

class ExtraColumnsFilter:

    def __init__(self, original_cols: list[str]):
        self.original_cols = set(original_cols)

    def filter_read(self, read: lqp_bundle.ir.Read) -> lqp_bundle.ir.Read:
        """Filter data_columns to only include columns in original_cols."""
        if not isinstance(read.read_type, lqp_bundle.ir.Export):
            return read

        config = read.read_type.config
        assert isinstance(config, lqp_bundle.ir.ExportCSVConfig) and config.data_columns is not None, \
            "Expected ExportCSVConfig with data_columns in the read type"

        data_columns = config.data_columns

        new_data_columns = [col for col in data_columns if col.column_name in self.original_cols]

        # Reconstruct the nested structure with filtered data_columns
        new_config = replace(config, data_columns=new_data_columns)
        new_read_type = replace(read.read_type, config=new_config)
        new_read = replace(read, read_type=new_read_type)

        return new_read



from relationalai.semantics.metamodel.rewrite import DNFUnionSplitter
from relationalai.semantics.sql.rewrite import ExtractNestedLogicals
from relationalai.semantics.lqp.rewrite import (
    ExtractKeys, Flatten, FunctionAnnotations, Splinter
)

__all__ = ["Splinter", "Flatten", "DNFUnionSplitter", "ExtractKeys",
           "ExtractNestedLogicals", "FunctionAnnotations"]



from relationalai.semantics import Model, Integer, define, select
from relationalai.experimental.paths.graph import Graph
from relationalai.experimental.paths.path_algorithms.two_sided_balls_repetition import two_balls_repetition


# Test with diamond graph
model_grid = Model("test_two_balls_repetition grid", dry_run=False)

grid = Graph.construct_grid(model_grid, 4)

n = Integer.ref()
u = grid.Node.ref()

source_1_1 = grid.Node.new(row = 1, col = 1)
source_1_2 = grid.Node.new(row = 2, col = 2)
target_1_1 = grid.Node.new(row = 3, col = 3)
target_1_2 = grid.Node.new(row = 4, col = 4)

Source_1 = model_grid.Concept("Source_1", extends=[grid.Node])
Target_1 = model_grid.Concept("Target_1", extends=[grid.Node])

define(Source_1(source_1_1))
define(Source_1(source_1_2))
define(Target_1(target_1_1))
define(Target_1(target_1_2))

source_ball_1, target_ball_1 = two_balls_repetition(grid, Source_1, Target_1, 1)
iter_source_1 = select(n, u.row, u.col).where(source_ball_1(n, u)).to_df()
set_source_ball_1 = set(row for row in iter_source_1.itertuples(index = False, name = None))
iter_target_1 = select(n, u.row, u.col).where(target_ball_1(n, u)).to_df()
set_target_ball_1 = set(row for row in iter_target_1.itertuples(index = False, name = None))

expected_source_ball_1 = {
    (0, 1, 1), (0, 2, 2)
}

expected_target_ball_1 = {
    (0, 4, 4), (1, 3, 4), (1, 4, 3), (0, 3, 3), (1, 3, 2), (1, 2, 3)
}

assert (
    set_source_ball_1 == expected_source_ball_1 and
    set_target_ball_1 == expected_target_ball_1
)


# First test with diamond graph
model_diamond = Model("test_two_balls_repetition diamond", dry_run=False)

diamond = Graph.construct_diamond(model_diamond, 3)

v = diamond.Node.ref()

source_2_1 = diamond.Node.new(id = 1)
source_2_2 = diamond.Node.new(id = 2)
target_2_1 = diamond.Node.new(id = 9)
target_2_2 = diamond.Node.new(id = 10)

Source_2 = model_diamond.Concept("Source_2", extends=[diamond.Node])
Target_2 = model_diamond.Concept("Target_2", extends=[diamond.Node])

define(Source_2(source_2_1))
define(Source_2(source_2_2))
define(Target_2(target_2_1))
define(Target_2(target_2_2))

source_ball_2, target_ball_2 = two_balls_repetition(diamond, Source_2, Target_2, 20)
iter_source_2 = select(n, v.id).where(source_ball_2(n, v)).to_df()
set_source_ball_2 = set(row for row in iter_source_2.itertuples(index = False, name = None))
iter_target_2 = select(n, v.id).where(target_ball_2(n, v)).to_df()
set_target_ball_2 = set(row for row in iter_target_2.itertuples(index = False, name = None))

expected_source_ball_2 = {
    (0, 1), (1, 2), (1, 3), (2, 4), (3, 5), (3, 6), (4, 7), (5, 8), (5, 9), (6, 10),
    (0, 2), (1, 4), (2, 5), (2, 6), (3, 7), (4, 8), (4, 9), (5, 10)
}

expected_target_ball_2 = {
    (0, 10), (1, 9), (1, 8), (2, 7), (3, 5), (3, 6), (4, 4), (5, 2), (5, 3), (6, 1),
    (0, 9), (1, 7), (2, 5), (2, 6), (3, 4), (4, 2), (4, 3), (5, 1)
}

assert (
    set_source_ball_2 == expected_source_ball_2 and
    set_target_ball_2 == expected_target_ball_2
)


# Second test with diamond graph
source_ball_3, target_ball_3 = two_balls_repetition(diamond, Source_2, Target_2, 0)
iter_source_3 = select(n, v.id).where(source_ball_3(n, v)).to_df()
set_source_ball_3 = set(row for row in iter_source_3.itertuples(index = False, name = None))
iter_target_3 = select(n, v.id).where(target_ball_3(n, v)).to_df()
set_target_ball_3 = set(row for row in iter_target_3.itertuples(index = False, name = None))

expected_source_ball_3 = {
    (0, 1), (0, 2)
}

expected_target_ball_3 = {
    (0, 10), (0, 9)
}

assert (
    set_source_ball_3 == expected_source_ball_3 and
    set_target_ball_3 == expected_target_ball_3
)


from __future__ import annotations
from typing import Tuple

from relationalai.semantics.metamodel import builtins, ir, factory as f, types, visitor, helpers
from relationalai.semantics.metamodel.compiler import Pass, group_tasks
from relationalai.semantics.metamodel.util import OrderedSet
from relationalai.semantics.metamodel.util import FrozenOrderedSet
from relationalai.semantics.metamodel.typer.typer import is_primitive

class FormatOutputs(Pass):
    def __init__(self, use_rel: bool=False):
        super().__init__()
        self._use_rel = use_rel

    #--------------------------------------------------
    # Public API
    #--------------------------------------------------
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        wide_outputs = options.get("wide_outputs", False)
        return self.OutputRewriter(wide_outputs, self._use_rel).walk(model)

    class OutputRewriter(visitor.Rewriter):
        def __init__(self, wide_outputs: bool = False, use_rel: bool = False):
            super().__init__()
            self.wide_outputs = wide_outputs
            self._use_rel = use_rel

        def handle_logical(self, node: ir.Logical, parent: ir.Node):
            # Rewrite children first
            node = super().handle_logical(node, parent)

            groups = group_tasks(node.body, {
                "outputs": ir.Output,
            })

            # If no outputs, return as is
            if not groups["outputs"]:
                return node

            if self.wide_outputs:
                return adjust_wide_outputs(node, groups["outputs"])

            return adjust_gnf_outputs(self._use_rel, node, groups["outputs"])

#--------------------------------------------------
# GNF vs wide output support
#--------------------------------------------------

# For wide outputs, only adjust the output task to include the keys.
# output looks like: (key0, key1, val0, val1, ...)
def adjust_wide_outputs(task: ir.Logical, outputs: OrderedSet[ir.Task]):
    body = list(task.body)
    for output in outputs:
        assert(isinstance(output, ir.Output))
        if output.keys:
            body.remove(output)
            body.append(rewrite_wide_output(output))
    return ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations)

# For GNF outputs we need to generate a rule for each "column" in the output
# and potentially one wide key column
def adjust_gnf_outputs(use_rel: bool, task: ir.Logical, outputs: OrderedSet[ir.Task]):
    body = list(task.body)
    for output in outputs:
        assert(isinstance(output, ir.Output))
        if output.keys:
            # Remove the original output. This is replaced by per-column outputs below
            body.remove(output)

            is_export = helpers.is_export(output)

            # Exports and Rel execution rely on all columns being in GNF format.
            if is_export or use_rel:
                _adjust_all_gnf_outputs(body, output, is_export)
            else: # Otherwise, put all keys into one wide keys relation
                _adjust_outputs_with_wide_keys(body, output)

    return ir.Logical(task.engine, task.hoisted, tuple(body), task.annotations)

# Generate an output for each "column"
# output looks like: def output(:cols, :col000, key0, key1, value)
def _adjust_all_gnf_outputs(body, output: ir.Output, is_export: bool):
    assert output.keys

    original_cols = OrderedSet()

    # Track existing column names (case-insensitively) to avoid value-column/key-column collisions
    # in exports, where Snowflake treats column names as case-insensitive.  e.g. value col "Person"
    # and key col "person" would collide as Snowflake column names.
    existing_names: set[str] = set()

    for idx, alias in enumerate(output.aliases):
        # Skip None values which are used as a placeholder for missing values
        if alias[1] is None:
            continue
        original_cols.add(alias[1])
        if is_export:
            existing_names.add(alias[0].lower())
        body.extend(_generate_output_column_gnf(output, idx, alias, is_export))

    idx = len(output.aliases)
    for key in output.keys:
        if key not in original_cols:
            col_name = key.name
            
            # If the key column name collides with an existing column name (case-insensitively),
            # append a "_key" suffix, then "_key_1", "_key_2", etc. until unique.
            # e.g. key "v" colliding with value col "v"  → "v_key"
            #      key "v" colliding with "v" and "v_key" → "v_key_1"
            if is_export:
                if col_name.lower() in existing_names:
                    col_name = f"{key.name}_key"
                    suffix = 1
                    while col_name.lower() in existing_names:
                        col_name = f"{key.name}_key_{suffix}"
                        suffix += 1
                existing_names.add(col_name.lower())

            body.extend(_generate_output_column_gnf(output, idx, (col_name, key), is_export))
            idx += 1

# Generate an output for each value "column" and one wide output for all the keys
#  * value output looks like: def output(:cols, :col000, key0, key1, value)
#  * key output looks like:
#    def output(:keys, output_key_0, output_key_1, other_key_0, ...)
#
# Exceptions: keys for exports and compound keys are converted to GNF, same as the
# value columns.
def _adjust_outputs_with_wide_keys(body, output: ir.Output):
    assert output.keys

    original_cols = OrderedSet()
    val_cols: list[Tuple[str, ir.Value] | None] = []
    key_cols: OrderedSet[Tuple[str, ir.Value]] = OrderedSet()
    key_cols.add(("keys", f.literal("keys", types.Symbol))) # name key col so we can identify it later
    for alias in output.aliases:
        # None values are used as a placeholder for missing values
        # They are added to maintain the correct col count when enumerated below
        if alias[1] is None:
            val_cols.append(None)
            continue

        original_cols.add(alias[1])

        if isinstance(alias[1], ir.Var) and alias[1] in output.keys: # note: skips compound keys
            key_cols.add(alias)
        else:
            val_cols.append(alias)

    # Add keys not in output to the end
    for key in output.keys:
        if key not in original_cols:
            key_cols.add((key.name, key))

    # Generate GNF val cols
    for idx, alias in enumerate(val_cols):
        if alias:
            new_col = _generate_output_column(output, idx, alias, key_cols)
            body.extend(new_col)

    # Create a wide key column with all keys
    if len(key_cols) > 1:
        body.append(ir.Output(
            output.engine,
            key_cols.frozen(),
            output.keys,
            output.annotations
        ))

# Generate a relation representing a single col in GNF form
def _generate_output_column(output: ir.Output, idx: int, alias: tuple[str, ir.Value], key_cols):
    if not output.keys:
        return [output]

    aliases = [("cols", f.literal("cols", types.Symbol))]
    aliases.append(("col", f.literal(f"col{idx:03}", types.Symbol)))

    # Append all keys at the start
    keys = iter(key_cols)
    assert next(keys) == ("keys", f.literal("keys", types.Symbol)) # skip col name
    for key in keys:
        aliases.append((f"key_{key[0]}_{idx}", key[1]))

    aliases.append(alias) # append val

    return [
        ir.Output(
            output.engine,
            FrozenOrderedSet.from_iterable(aliases),
            output.keys,
            output.annotations
        )
    ]

# Generate a relation representing a single col in GNF form for export
def _generate_output_column_gnf(output: ir.Output, idx: int, alias: tuple[str, ir.Value], is_export: bool):
    if not output.keys:
        return [output]

    aliases = [("cols", f.literal("cols", types.Symbol))] if not is_export else []
    aliases.append(("col", f.literal(f"col{idx:03}", types.Symbol)))

    for k in output.keys:
        aliases.append((f"key_{k.name}_{idx}", k))

    if (is_export and
        isinstance(alias[1], ir.Var) and
        (not is_primitive(alias[1].type) or alias[1].type == types.Hash)):

        # Choose the base name for the uuid variable:
        #  - If it is a key column: use alias[1].name (original variable name) to keep the uuid name
        #    clean, unaffected by collision-avoidance renaming on alias[0].
        #    e.g. key "v" renamed to "v_key" → uuid stays "v_3_uuid", not "v_key_3_uuid"
        #  - If it is a Hash-typed value column: use alias[0] (user-provided column name) so the
        #    uuid export column remains meaningful to the user.
        #    e.g. value col "my_hash" → uuid is "my_hash_2_uuid"
        uuid_base = alias[1].name if alias[1] in output.keys else alias[0]
        uuid = f.var(f"{uuid_base}_{idx}_uuid", types.String)

        if not is_primitive(alias[1].type):
            # For non-primitive types, we keep the original alias
            aliases.append((alias[0], uuid))
        else:
            # For Hash types, we use the uuid name as alias
            aliases.append((uuid.name, uuid))

        return [
            ir.Lookup(None, builtins.uuid_to_string, (alias[1], uuid)),
            ir.Output(
                output.engine,
                FrozenOrderedSet.from_iterable(aliases),
                output.keys,
                output.annotations
            )
        ]
    else:
        aliases.append(alias)

        return [
            ir.Output(
                output.engine,
                FrozenOrderedSet.from_iterable(aliases),
                output.keys,
                output.annotations
            )
        ]

def rewrite_wide_output(output: ir.Output):
    assert(output.keys)

    # Only append keys that are not already in the output
    suffix_keys = []
    for key in output.keys:
        if all([val is not key for _, val in output.aliases]):
            suffix_keys.append(key)

    aliases: OrderedSet[Tuple[str, ir.Value]] = OrderedSet()

    # Add the remaining args, unless it is already a key
    for name, val in output.aliases:
        if not isinstance(val, ir.Var) or val not in suffix_keys:
            aliases.add((name, val))

    # Add the keys to the output
    for key in suffix_keys:
        aliases.add((key.name, key))

    # TODO - we are assuming that the Rel compiler will translate nullable lookups
    # properly, returning a `Missing` if necessary, like this:
    # (nested_192(_adult, _adult_name) or (not nested_192(_adult, _) and _adult_name = Missing)) and
    return ir.Output(
        output.engine,
        aliases.frozen(),
        output.keys,
        output.annotations
    )

    # TODO: in the rel compiler, see if we can do this outer join
    # 1. number of keys
    # 2. each relation
    # 3. each variable, starting with the keys
    # 4. tag output with @arrow

    # @arrow def output(_book, _book_title, _author_name):
    #   rel_primitive_outer_join(#1, book_title, author_name, _book, _book_title, _author_name)
    # def output(p, n, c):
    #     rel_primitive_outer_join(#1, name, coolness, p, n, c)



from relationalai.semantics.lqp import lqp_bundle
from relationalai.semantics.metamodel import ir, builtins

from relationalai.semantics.metamodel.util import FrozenOrderedSet

from dataclasses import dataclass
from hashlib import sha256
from typing import Tuple

class UniqueNames:
    def __init__(self):
        # Track count of seen names
        self.seen = dict[str, int]()
        # Maps id to unique name
        self.id_to_name = dict[int,str]()

    def get_name(self, name: str) -> str:
        # Names are used as-is for display in orig_names (Python-qualified format).
        # Collision handling below ensures uniqueness without sanitization.
        if name not in self.seen:
            self.seen[name] = 1
            return f"{name}"

        self.seen[name] += 1
        id = self.seen[name]
        # If the original name has a suffix we can get collisions with generated names,
        # so test the new name.
        while f"{name}_{id}" in self.seen:
            id += 1
            self.seen[name] = id
        new_name = f"{name}_{id}"
        self.seen[new_name] = 1
        return new_name

    # Get a unique name for the given id. If the id is already in the map, return the
    # existing name. Otherwise, generate a new name using the suggested_name and
    # store it in the map.
    def get_name_by_id(self, id: int, suggested_name:str) -> str:
        if id in self.id_to_name:
            return self.id_to_name[id]

        name = self.get_name(suggested_name)
        self.id_to_name[id] = name
        return name

@dataclass(frozen=True)
class ExportDescriptor:
    relation_id: lqp_bundle.ir.RelationId
    column_name: str
    column_number: int
    column_type: lqp_bundle.ir.Type
    export_fqn: str

# Type alias for export_info structure: list of (export_fqn, filename, column_types)
ExportInfo = list[tuple[str, str, list]]

def extract_export_fqn(annotations: list[ir.Annotation] | FrozenOrderedSet[ir.Annotation]) -> str:
    """
    Extract the export FQN from annotations.

    Returns the export FQN string if an export annotation is found, "unnamed_table" otherwise.
    """
    for annotation in annotations:
        if annotation.relation == builtins.export:
            if annotation.args:
                export_fqn_literal = annotation.args[0]
                if isinstance(export_fqn_literal, ir.Literal) and isinstance(export_fqn_literal.value, str):
                    return export_fqn_literal.value
    return "unnamed_table"

class TranslationCtx:
    def __init__(self, def_names: UniqueNames = UniqueNames()):
        # TODO: comment these fields
        self.def_names = def_names
        self.var_names = UniqueNames()
        self.output_names = UniqueNames()
        # A counter for break rules generated during translation of while loops
        self.break_rule_counter = 0
        # Map relation IDs to their original names for debugging and pretty printing.
        self.rel_id_to_orig_name = {}
        self.output_ids: list[tuple[lqp_bundle.ir.RelationId, str]] = []
        self.export_descriptors: list[ExportDescriptor] = []

def gen_rel_id(ctx: TranslationCtx, orig_name: str, suffix: str = "") -> lqp_bundle.ir.RelationId:
    relation_id = lqp_bundle.ir.RelationId(id=lqp_hash(orig_name + suffix), meta=None)
    ctx.rel_id_to_orig_name[relation_id] = orig_name
    return relation_id

def gen_unique_var(ctx: TranslationCtx, name_hint: str) -> lqp_bundle.ir.Var:
    """
    Generate a new variable with a unique name based on the provided hint.
    """
    name = ctx.var_names.get_name(name_hint)
    return lqp_bundle.ir.Var(name=name, meta=None)

def is_constant(arg, expected_type):
    """
    Check if the argument is a constant of the expected type.
    """
    if isinstance(arg, ir.Literal):
        return is_constant(arg.value, expected_type)

    return isinstance(arg, expected_type)

def rename_vars_var(var: lqp_bundle.ir.Var, var_map: dict[str, lqp_bundle.ir.Var]) -> lqp_bundle.ir.Var:
    return var_map.get(var.name, var)

def rename_vars_relterm(term: lqp_bundle.ir.RelTerm, var_map: dict[str, lqp_bundle.ir.Var]) -> lqp_bundle.ir.RelTerm:
    if isinstance(term, lqp_bundle.ir.Var):
        return rename_vars_var(term, var_map)
    else:
        return term  # Constants do not change

def rename_vars_term(term: lqp_bundle.ir.Term, var_map: dict[str, lqp_bundle.ir.Var]) -> lqp_bundle.ir.Term:
    if isinstance(term, lqp_bundle.ir.Var):
        return rename_vars_var(term, var_map)
    else:
        return term  # Constants do not change

def rename_vars_abstraction(abstraction: lqp_bundle.ir.Abstraction, var_map: dict[str, lqp_bundle.ir.Var]) -> lqp_bundle.ir.Abstraction:
    new_vars = [(var_map.get(var.name, var), typ) for (var, typ) in abstraction.vars]
    new_value = rename_vars_formula(abstraction.value, var_map)
    return lqp_bundle.ir.Abstraction(vars=new_vars, value=new_value, meta=abstraction.meta)

def rename_vars_formula(formula: lqp_bundle.ir.Formula, var_map: dict[str, lqp_bundle.ir.Var]) -> lqp_bundle.ir.Formula:
    if isinstance(formula, lqp_bundle.ir.Primitive):
        return lqp_bundle.ir.Primitive(
            name=formula.name,
            terms=[rename_vars_relterm(term, var_map) for term in formula.terms],
            meta=formula.meta
        )
    elif isinstance(formula, lqp_bundle.ir.Atom):
        return lqp_bundle.ir.Atom(
            name=formula.name,
            terms=[rename_vars_term(term, var_map) for term in formula.terms],
            meta=formula.meta
        )
    elif isinstance(formula, lqp_bundle.ir.RelAtom):
        return lqp_bundle.ir.RelAtom(
            name=formula.name,
            terms=[rename_vars_relterm(term, var_map) for term in formula.terms],
            meta=formula.meta
        )
    elif isinstance(formula, lqp_bundle.ir.Not):
        return lqp_bundle.ir.Not(arg=rename_vars_formula(formula.arg, var_map), meta=formula.meta)
    elif isinstance(formula, lqp_bundle.ir.Exists):
        return lqp_bundle.ir.Exists(body=rename_vars_abstraction(formula.body, var_map), meta=None)
    elif isinstance(formula, lqp_bundle.ir.Reduce):
        return lqp_bundle.ir.Reduce(
            op=formula.op,
            body=rename_vars_abstraction(formula.body, var_map),
            terms=[rename_vars_term(term, var_map) for term in formula.terms],
            meta=formula.meta
        )
    elif isinstance(formula, lqp_bundle.ir.FFI):
        return lqp_bundle.ir.FFI(
            meta=formula.meta,
            name=formula.name,
            args=[rename_vars_abstraction(arg, var_map) for arg in formula.args],
            terms=[rename_vars_term(term, var_map) for term in formula.terms]
        )
    elif isinstance(formula, lqp_bundle.ir.Conjunction):
        return lqp_bundle.ir.Conjunction(
            args=[rename_vars_formula(arg, var_map) for arg in formula.args],
            meta=formula.meta
        )
    elif isinstance(formula, lqp_bundle.ir.Disjunction):
        return lqp_bundle.ir.Disjunction(
            args=[rename_vars_formula(arg, var_map) for arg in formula.args],
            meta=formula.meta
        )
    elif isinstance(formula, lqp_bundle.ir.Cast):
        return lqp_bundle.ir.Cast(
            input=rename_vars_term(formula.input, var_map),
            result=rename_vars_term(formula.result, var_map),
            meta=formula.meta,
        )
    else:
        raise NotImplementedError(f"Unknown formula type: {type(formula)}")

def lqp_hash(node: str) -> int:
    h = int.from_bytes(sha256(node.encode()).digest(), byteorder='big', signed=False)
    # Ensure it's within the 128-bit range
    return h % (2**128)

def output_names(aliases: FrozenOrderedSet[Tuple[str, ir.Value]]) -> list[str]:
    return [v[0] for v in aliases]



from relationalai.semantics.metamodel.compiler import Pass
from relationalai.semantics.metamodel import ir, builtins as rel_builtins, factory as f, visitor
from relationalai.semantics.metamodel import types

# Generate date arithmetic expressions, such as
# `rel_primitive_date_add(:day, [date] delta, res_2)` by finding the period
# expression for the delta and adding the period type to the date arithmetic expression.
#
# date_add and it's kin are generated by a period expression, e.g.,
# `day(delta, res_1)`
# followed by the date arithmetic expression using the period
# `date_add([date] res_1 res_2)`
class PeriodMath(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        period_rewriter = self.PeriodRewriter()
        model = period_rewriter.walk(model)
        period_math_rewriter = self.PeriodMathRewriter(period_rewriter.period_vars)
        model = period_math_rewriter.walk(model)
        return model

    # Find all period builtins. We need to make them safe for the emitter (either by
    # translating to a cast, or removing) and store the variable and period type for use
    # in the date/datetime add/subtract expressions.
    class PeriodRewriter(visitor.Rewriter):
        def __init__(self):
            super().__init__()
            self.period_vars: dict[ir.Var, str] = {}

        def handle_lookup(self, node: ir.Lookup, parent: ir.Node) -> ir.Lookup:
            if not rel_builtins.is_builtin(node.relation):
                return node

            if node.relation.name not in {
                "year", "month", "week", "day", "hour", "minute", "second", "millisecond", "microsecond", "nanosecond"
            }:
                return node

            assert len(node.args) == 2, "Expect 2 arguments for period builtins"
            assert isinstance(node.args[1], ir.Var), "Expect result to be a variable"
            period = node.relation.name
            result_var = node.args[1]
            self.period_vars[result_var] = period

            # Ideally we could now remove the unused and unhandled period type construction
            # but we may also need to cast the original variable to an Int64 for use by the
            # date/datetime add/subtract expressions.
            # TODO: Remove the node entirely where possible and update uses of the result
            return f.lookup(rel_builtins.cast, [types.Int64, node.args[0], result_var])

    # Update date/datetime add/subtract expressions with period information.
    class PeriodMathRewriter(visitor.Rewriter):
        def __init__(self, period_vars: dict[ir.Var, str]):
            super().__init__()
            self.period_vars: dict[ir.Var, str] = period_vars

        def handle_lookup(self, node: ir.Lookup, parent: ir.Node) -> ir.Lookup:
            if not rel_builtins.is_builtin(node.relation):
                return node

            if node.relation.name in { "date_diff", "datetime_diff" }:
                # ensure the part argument is a symbol literal for the emitter
                part = node.args[0]
                assert isinstance(part, ir.Literal)
                return f.lookup(node.relation, [f.literal(part.value, types.Symbol), node.args[1], node.args[2], node.args[3]])

            if node.relation.name not in {
                "date_add", "date_subtract", "datetime_add", "datetime_subtract"
            }:
                return node

            if len(node.args) == 4:
                # We've already visited this lookup
                return node

            assert isinstance(node.args[1], ir.Var), "Expect period to be a variable"
            period_var = node.args[1]
            assert period_var in self.period_vars, "datemath found, but no vars to insert"

            period = self.period_vars[period_var]

            new_args = [f.literal(period, types.Symbol)] + [arg for arg in node.args]

            return f.lookup(node.relation, new_args)



"""
Constructing Metamodel IR with Algorithms

We introduce a set of programmatic constructs that provide a convenient syntax for
constructing PyRel's metamodel IR representations for Loopy algorithms. Importantly, these
macros construct a new model using PyRel declarations constructed with a _base model_. The
base model needs to be also used to declare all concepts and relationships.

Below we illustrate the use of these macros by constructing a simple reachability
algorithm, whose Rel-like pseudo-code is as follows:

```
algorithm
    setup
        def edge = { (1,2); (2,3); (3,4) }
        def source = { 1 }
    end setup
    @global empty reachable = {}
    loop
        def frontier = source
        def reachable = frontier
        while (true)
            def next_frontier = frontier . edge
            def frontier = next_frontier
            monus frontier = reachable    # frontier = frontier - reachable
            upsert reachable = frontier   # reachable = reachable ∪ frontier
            break break_reachable = empty(frontier)
        end while
    end loop
end algorithm
```

The PyRel's metamodel IR for the above algorithm is constructed with the utilities as
follows.

```
base_model = Model("algorithm_builder", dry_run=True)

# Input (context) data

edge = base_model.Relationship("Edge from {source:int} to {target:int}")
source = base_model.Relationship("Source node {node:int}")

with algorithm(base_model):
    setup(
        define(edge(1,2), edge(2,3), edge(3,4), edge(4,1))),
        define(source(1))
    )

    # "local" variables and relations
    n = Integer.ref()
    m = Integer.ref()
    reachable = base_model.Relationship("Reachable node {node:int}")
    frontier = base_model.Relationship("Frontier node {node:int}")
    next_frontier = base_model.Relationship("Next frontier node {node:int}")

    global_(empty(define(reachable(n))))
    assign(define(frontier(n)).where(source(n)))
    assign(define(reachable(n)).where(frontier(n)))
    with while_():
        assign(define(next_frontier(m)).where(frontier(n), edge(n, m)))
        assign(define(frontier(m)).where(next_frontier(m)))
        monus(define(frontier(n)).where(reachable(n)))
        upsert(0)(define(reachable(n)).where(frontier(n)))
        break_(where(not_(frontier(n))))

# Prints the PyRel Metamodel (IR)
print(get_metamodel())

# Prints the LQP transaction
print(get_lqp_str())
```
"""
from relationalai.semantics import Model
from relationalai.semantics.metamodel import factory, ir, types
from relationalai.semantics.internal.internal import Fragment
from relationalai.semantics.lqp.algorithms import (
    mk_empty, mk_assign, mk_upsert, mk_global, mk_monus
)
from relationalai.semantics.lqp.constructors import mk_transaction
from relationalai.semantics.lqp.compiler import Compiler
from relationalai.semantics.lqp import lqp_bundle
from relationalai.semantics.lqp import builtins
from typing import cast, TypeGuard, Optional, Sequence
from lqp.gen.pretty import pretty
import threading
from contextlib import contextmanager


# While the constructors are very light-weight they enforce
# the following grammar for algorithms:
#
# <Algorithm> := with algorithm(base_model): <Script>
# <Script> := <Instruction>*
# <Instruction> := <BaseInstruction> | <Loop>
# <BaseInstruction> := [global_(] empty(Fragment) [)]
#                    | [global_(] assign(<Fragment>) [)]
#                    | break(<Fragment>)
#                    | upsert(<Int>)(<Fragment>)
#                    | monus(<Fragment>)
# <Loop> := with while_(): <Script>
#
# Note: global_ annotation can only be used on top-level empty and assign instructions at the
# top-level of the algorithm script.

_storage = threading.local()

def get_builder() -> 'AlgorithmBuilder':
    """ Retrieves the thread-local AlgorithmBuilder instance."""
    global _storage
    if not(hasattr(_storage, "algorithm_builder")):
        _storage.algorithm_builder = AlgorithmBuilder()
    return _storage.algorithm_builder

def get_metamodel() -> ir.Model:
    """ Retrieves the compiled metamodel IR for the previous algorithm. Can only be used
    after an algorithm has been defined."""
    return get_builder().get_metamodel()

def get_lqp_str() -> str:
    """ Retrieves the LQP string representation for the previous algorithm. Can only be used
    after an algorithm has been defined."""
    return get_builder().get_lqp_str()

def get_lqp():
    """ Retrieves the LQP string representation for the previous algorithm. Can only be used
    after an algorithm has been defined."""
    return get_builder().get_lqp()

@contextmanager
def algorithm(model:Model):
    """ Context manager for defining an algorithm on the given base model."""
    get_builder().begin_algorithm(model)
    yield
    get_builder().end_algorithm()

@contextmanager
def while_():
    """ Context manager for defining a while loop within an algorithm."""
    get_builder().begin_while_loop()
    yield
    get_builder().end_while_loop()

def setup(*stmts:Fragment):
    """ Defines the setup section of an algorithm: a collection of PyRel statement that
    prepare input data for the algorithm."""
    builder = get_builder()
    assert len(builder.script_stacks) == 1, "setup can only be called at the top-level of an algorithm"
    assert builder.setup_fragments is None, "setup can only be called once per algorithm"
    builder.set_setup_fragments(stmts)

def global_(pos:int):
    """ Marks a top-level `empty` or `assign` instruction as defining a global relation."""
    assert type(pos) is int, "global_ can only be applied to empty and assign"
    builder = get_builder()
    assert len(builder.script_stacks) == 1, "global_ can only be applied to top-level instructions"
    assert len(builder.script_stacks[0].instructions) == pos + 1
    task = cast(ir.Task, mk_global(builder.script_stacks[0].instructions[pos]))
    builder.script_stacks[0].instructions[pos] = task
    builder.add_global_relation(task)

def empty(stmt) -> int:
    """ Marks a PyRel statement as an assignment of empty relation. The statement must not
    have a body (no where clause)."""
    assert has_empty_body(stmt), "Empty instruction must have an empty body"
    task = get_builder().compile_statement(stmt)
    task = cast(ir.Task, mk_empty(task))
    return get_builder().append_task(task)

def assign(stmt) -> int:
    """ Marks a PyRel statement as an assignment instruction."""
    task = get_builder().compile_statement(stmt)
    task = cast(ir.Task, mk_assign(task))
    return get_builder().append_task(task)

def upsert_with_arity(arity:int, stmt:Fragment):
    task = get_builder().compile_statement(stmt)
    task = cast(ir.Task, mk_upsert(task, arity))
    get_builder().append_task(task)

def upsert(arity:int):
    """ Marks a PyRel statement as an upsert instruction with the given arity."""
    assert type(arity) is int and arity >= 0, "arity must be a non-negative integer"
    return lambda stmt: upsert_with_arity(arity, stmt)

def monus(stmt: Fragment) -> int:
    """ Marks a PyRel statement as a Boolean monus (set difference) instruction."""
    task = get_builder().compile_statement(stmt)
    task = cast(ir.Task, mk_monus(task, types.Bool, "or", 0))
    return get_builder().append_task(task)

def break_(stmt, debug_anno=False):
    """ Marks a PyRel statement as a break instruction. The statement must be headless (no define clause)."""
    assert has_no_head(stmt), "Break instruction must have a headless fragment"
    task = get_builder().compile_statement(stmt)
    assert isinstance(task, ir.Logical)
    break_condition = [cond for cond in task.body if not isinstance(cond, ir.Update)]
    annos = [builtins.relation_content_debug_annotation(1)] if debug_anno else []
    break_node = factory.break_(factory.logical(break_condition), annos=annos)
    get_builder().append_task(break_node)

def has_empty_body(stmt) -> TypeGuard[Fragment]:
    if not isinstance(stmt, Fragment):
        return False
    return len(stmt._where) == 0

def has_no_head(frag):
    return len(frag._define) == 0


class ScriptBuilder:
    """
    Builder for Loopy scripts.
    """
    def __init__(self):
        self.instructions:list[ir.Task] = []

    def add_task(self, instr:ir.Task) -> int:
        self.instructions.append(instr)
        return len(self.instructions) - 1

    def build_script(self, annos:list[ir.Annotation]) -> ir.Sequence:
        return factory.sequence(
            tasks=self.instructions,
            annos=[builtins.script_annotation()] + annos
        )


class AlgorithmBuilder:
    """
    Builder for Loopy algorithms.
    """
    def __init__(self):
        self.script_stacks:list[ScriptBuilder] = []
        self.compiled_model:Optional[ir.Model] = None
        self.global_relations:list[str] = []
        self.base_model:Optional[Model] = None
        self.setup_fragments:Optional[list[Fragment]] = None

    def begin_algorithm(self, base_model:Model):
        self.base_model = base_model
        self.script_stacks = [ScriptBuilder()]
        self.compiled_model = None
        self.global_relations:list[str] = []
        self.setup_fragments:Optional[list[Fragment]] = None

    def add_global_relation(self, task:ir.Task):
        assert isinstance(task, ir.Logical)
        for t in task.body:
            if isinstance(t, ir.Update):
                if t.relation.name not in self.global_relations:
                    self.global_relations.append(t.relation.name)

    def set_setup_fragments(self, fragments:Sequence[Fragment]):
        self.setup_fragments = list(fragments)

    def compile_statement(self, stmt:Fragment) -> ir.Task:
        assert self.base_model is not None
        task = self.base_model._compiler.compile_task(stmt)
        return task

    def append_task(self, task:ir.Task) -> int:
        assert len(self.script_stacks) > 0
        return self.script_stacks[-1].add_task(task)

    def begin_while_loop(self):
        script_builder = ScriptBuilder()
        self.script_stacks.append(script_builder)

    def end_while_loop(self):
        script_builder = self.script_stacks.pop()
        while_script = script_builder.build_script([builtins.while_annotation()])
        loop = factory.loop(while_script, annos=[builtins.while_annotation()])
        self.append_task(loop)

    def end_algorithm(self):
        assert len(self.script_stacks) == 1
        script_builder = self.script_stacks.pop()
        algorithm_script = script_builder.build_script([builtins.algorithm_annotation()])
        setup = self.compile_setup()
        algorithm_logical = factory.logical(setup + [algorithm_script])
        self.compiled_model = factory.compute_model(algorithm_logical)

    def compile_setup(self) -> list[ir.Logical]:
        if self.setup_fragments is None:
            return []
        assert self.setup_fragments is not None
        assert self.base_model is not None
        setup_tasks = []
        for stmt in self.setup_fragments:
            task = self.base_model._compiler.compile_task(stmt)
            setup_tasks.append(task)
        return setup_tasks

    def get_metamodel(self) -> ir.Model:
        """ Retrieves the compiled metamodel IR for the previous algorithm. """
        metamodel = self.compiled_model
        assert metamodel is not None, "No metamodel available. You must first define algorithm."
        return metamodel

    def get_lqp_str(self) -> str:
        lqp = self.get_lqp()
        return pretty(lqp_bundle.ir_to_proto(lqp))

    def get_lqp(self):
        model = self.get_metamodel()

        compiler = Compiler()
        rewritten_model = compiler.rewrite(model)
        write_epoch = compiler.do_compile(rewritten_model, {'fragment_id': b"f1"})[1]

        define = cast(lqp_bundle.ir.Define, write_epoch.writes[0].write_type)
        debug_info = define.fragment.debug_info

        read_epoch = self._build_read_epoch(debug_info)

        transaction = mk_transaction([write_epoch, read_epoch])

        return transaction

    def _build_read_epoch(self, debug_info:lqp_bundle.ir.DebugInfo) -> lqp_bundle.ir.Epoch:
        reads = []

        relation_id:dict[str,lqp_bundle.ir.RelationId] = dict()
        for rel_id, rel_name in debug_info.id_to_orig_name.items():
            if rel_name in self.global_relations:
                relation_id[rel_name] = rel_id

        global_relation_names = [rel for rel in self.global_relations if rel in relation_id]

        for (i, rel_name) in enumerate(global_relation_names):
            read = lqp_bundle.ir.Read(
                meta = None,
                read_type = lqp_bundle.ir.Output(
                    meta=None,
                    name=f"{rel_name}",
                    relation_id=relation_id[rel_name],
                )
            )
            reads.append(read)

        read_epoch = lqp_bundle.ir.Epoch(
            meta = None,
            writes = [],
            reads = reads,
        )

        return read_epoch



"""Optimization and constraint programming solver interfaces.

This package provides solver model interfaces for defining and solving
mathematical optimization and constraint programming problems using
RelationalAI's solver infrastructure.
"""

from __future__ import annotations

from .solver import Provider, Solver

from .common import all_different, implies, make_name, special_ordered_set_type_2
from .solver_model import SolverModel

__version__ = "0.0.0"

__all__ = [
    "Solver",
    "Provider",
    "SolverModel",
    "make_name",
    "all_different",
    "implies",
    "special_ordered_set_type_2",
]



from .denormalize import Denormalize
from .extract_nested_logicals import ExtractNestedLogicals
from .flatten import Flatten
from .recursive_union import RecursiveUnion
from .double_negation import DoubleNegation
from .sort_output_query import SortOutputQuery

__all__ = ["Denormalize", "ExtractNestedLogicals", "Flatten", "RecursiveUnion", "DoubleNegation", "SortOutputQuery"]
